"""
Copyright 2021, CCL Forensics
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import pathlib
import typing
import dataclasses
import re
import collections.abc as col_abc
from types import MappingProxyType

from .storage_formats import ccl_leveldb
from .common import KeySearch

__version__ = "0.6"
__description__ = "Module for reading the Chromium leveldb sessionstorage format"
__contact__ = "Alex Caithness"

# See: https://source.chromium.org/chromium/chromium/src/+/main:components/services/storage/dom_storage/session_storage_metadata.cc
# et al

_NAMESPACE_PREFIX = b"namespace-"
_MAP_ID_PREFIX = b"map-"

log = None


@dataclasses.dataclass(frozen=True)
class SessionStoreValue:
    host: typing.Optional[str]
    key: str
    value: str
    # guid: typing.Optional[str]
    leveldb_sequence_number: int
    is_deleted: bool = False

    @property
    def record_location(self) -> str:
        return f"Leveldb Seq: {self.leveldb_sequence_number}"


class SessionStoreDb:
    # todo: get all grouped by namespace by host?
    # todo: get all grouped by namespace by host.key?
    # todo: consider refactoring to only getting metadata on first pass and everything else on demand?
    def __init__(self, in_dir: pathlib.Path):
        if not in_dir.is_dir():
            raise IOError("Input directory is not a directory")

        self._ldb = ccl_leveldb.RawLevelDb(in_dir)

        # If performance is a concern we should refactor this, but slow and steady for now

        # First collect the namespace (session/tab guid  + host) and map-ids together
        self._map_id_to_host = {}  # map_id: host
        self._deleted_keys = set()

        for rec in self._ldb.iterate_records_raw():
            if rec.user_key.startswith(_NAMESPACE_PREFIX):
                if rec.user_key == _NAMESPACE_PREFIX:
                    continue  # bogus entry near the top usually
                try:
                    key = rec.user_key.decode("utf-8")
                except UnicodeDecodeError:
                    print(f"Invalid namespace key: {rec.user_key}")
                    continue

                split_key = key.split("-", 2)
                if len(split_key) != 3:
                    print(f"Invalid namespace key: {key}")
                    continue

                _, guid, host = split_key

                if not host:
                    continue  # TODO investigate why this happens

                # normalize host to lower just in case
                host = host.lower()
                guid_host_pair = guid, host

                if rec.state == ccl_leveldb.KeyState.Deleted:
                    self._deleted_keys.add(guid_host_pair)
                else:
                    try:
                        map_id = rec.value.decode("utf-8")
                    except UnicodeDecodeError:
                        print(f"Invalid namespace value: {key}")
                        continue

                    if not map_id:
                        continue  # TODO: investigate why this happens/do we want to keep the host around somewhere?

                    #if map_id in self._map_id_to_host_guid and self._map_id_to_host_guid[map_id] != guid_host_pair:
                    if map_id in self._map_id_to_host and self._map_id_to_host[map_id] != host:
                        print("Map ID Collision!")
                        print(f"map_id: {map_id}")
                        print(f"Old host: {self._map_id_to_host[map_id]}")
                        print(f"New host: {guid_host_pair}")
                        raise ValueError("map_id collision")
                    else:
                        self._map_id_to_host[map_id] = host

        # freeze stuff
        self._map_id_to_host = MappingProxyType(self._map_id_to_host)

        self._deleted_keys = frozenset(self._deleted_keys)
        self._deleted_keys_lookup: dict[str, tuple] = {}

        self._host_lookup = {}  # {host: {ss_key: [SessionStoreValue, ...]}}
        self._orphans = []  #  list of tuples of key, value where we can't get the host
        for rec in self._ldb.iterate_records_raw():
            if rec.user_key.startswith(_MAP_ID_PREFIX):
                try:
                    key = rec.user_key.decode("utf-8")
                except UnicodeDecodeError:
                    print(f"Invalid map id key: {rec.user_key}")
                    continue

                # if rec.state == ccl_leveldb.KeyState.Deleted:
                #     continue  # TODO: do we want to keep the key around because the presence is important?

                split_key = key.split("-", 2)
                if len(split_key) != 3:
                    print(f"Invalid map id key: {key}")
                    continue

                _, map_id, ss_key = split_key

                if not split_key:
                    # TODO what does it mean when there is no key here?
                    #      The value will also be a single number (encoded utf-8)
                    continue

                try:
                    value = rec.value.decode("UTF-16-LE") if rec.state == ccl_leveldb.KeyState.Live else None
                except UnicodeDecodeError:
                    print(f"Error decoding value for {key}")
                    print(f"Raw Value: {rec.value}")
                    continue

                host = self._map_id_to_host.get(map_id)
                if not host:
                    self._orphans.append(
                        (ss_key,
                         SessionStoreValue(None, ss_key, value, rec.seq, rec.state == ccl_leveldb.KeyState.Deleted)
                         ))
                else:
                    self._host_lookup.setdefault(host, {})
                    self._host_lookup[host].setdefault(ss_key, [])
                    self._host_lookup[host][ss_key].append(
                        SessionStoreValue(host, ss_key, value, rec.seq, rec.state == ccl_leveldb.KeyState.Deleted))

    def __contains__(self, item: typing.Union[str, typing.Tuple[str, str]]) -> bool:
        """
        :param item: either the host as a str or a tuple of the host and a key (both str)
        :return: if item is a str, returns true if that host is present, if item is a tuple of (str, str), returns True
            if that host and key pair are present
        """

        if isinstance(item, str):
            return item in self._host_lookup
        elif isinstance(item, tuple) and len(item) == 2:
            host, key = item
            return host in self._host_lookup and key in self._host_lookup[host]
        else:
            raise TypeError("item must be a string or a tuple of (str, str)")

    def iter_hosts(self) -> typing.Iterable[str]:
        """
        :return: yields the hosts present in this SessionStorage
        """
        yield from self._host_lookup.keys()

    def get_all_for_host(self, host: str) -> dict[str, tuple[SessionStoreValue, ...]]:
        """
        DEPRECATED
        :param host: the host (domain name) for the session storage
        :return: a dictionary where the keys are storage keys and the values are tuples of SessionStoreValue objects
            for that key. Multiple values may be returned as deleted or old values may be recovered.
        """
        if host not in self:
            return {}
        result_raw = dict(self._host_lookup[host])
        for ss_key in result_raw:
            result_raw[ss_key] = tuple(result_raw[ss_key])
        return result_raw

    def _search_host(self, host: KeySearch) -> list[str]:
        if isinstance(host, str):
            return [host]
        elif isinstance(host, re.Pattern):
            return [x for x in self._host_lookup if host.search(x)]
        elif isinstance(host, col_abc.Collection):
            return list(set(host) & self._host_lookup.keys())
        elif isinstance(host, col_abc.Callable):
            return [x for x in self._host_lookup if host(x)]
        else:
            raise TypeError(f"Unexpected type: {type(host)} (expects: {KeySearch})")

    def iter_records_for_host(
            self, host: KeySearch, *,
            include_deletions=False, raise_on_no_result=True) -> col_abc.Iterable[SessionStoreValue]:
        """
        :param host: storage key (host) for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string (each host) and returns a bool.
        :param include_deletions: if True, records related to deletions will be included
        :param raise_on_no_result: if True (the default) if no matching storage keys are found, raise a KeyError
        (these will have None as values).
        :return: iterable of SessionStoreValue
        """
        if isinstance(host, str):
            if raise_on_no_result and host not in self._host_lookup:
                raise KeyError(host)
            for records in self._host_lookup[host].values():
                for rec in records:
                    if include_deletions or not rec.is_deleted:
                        yield rec
        elif isinstance(host, re.Pattern) or isinstance(host, col_abc.Collection) or isinstance(host, col_abc.Callable):
            found_hosts = self._search_host(host)
            if raise_on_no_result and not found_hosts:
                raise KeyError(host)
            for found_host in found_hosts:
                for records in self._host_lookup[found_host].values():
                    for rec in records:
                        if include_deletions or not rec.is_deleted:
                            yield rec
        else:
            raise TypeError(f"Unexpected type for host: {type(host)} (expects: {KeySearch})")

    def iter_all_records(self, *, include_deletions=False, include_orphans=False):
        """
        Returns all records recovered from session storage
        :param include_deletions: if True, records related to deletions will be included
        :param include_orphans: if True, records which cannot be associated with a host will be included
        """
        for host in self.iter_hosts():
            yield from self.iter_records_for_host(host, include_deletions=include_deletions)
        if include_orphans:
            yield from (x[1] for x in self.iter_orphans())

    def get_session_storage_key(self, host: str, key: str) -> tuple[SessionStoreValue, ...]:
        """
        DEPRECATED
        :param host: the host (domain name) for the session storage
        :param key: the storage key
        :return: a tuple of SessionStoreValue matching the host and key. Multiple values may be returned as deleted or
            old values may be recovered.
        """
        if (host, key) not in self:
            return tuple()
        return tuple(self._host_lookup[host][key])

    def iter_records_for_session_storage_key(
            self, host: KeySearch, key: KeySearch, *,
            include_deletions=False, raise_on_no_result=True) -> col_abc.Iterable[SessionStoreValue]:
        """
        :param host: storage key (host) for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string (each host) and returns a bool.
        :param key: script defined key for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool.
        :param include_deletions: if True, records related to deletions will be included
        :param raise_on_no_result: if True (the default) if no matching storage keys are found, raise a KeyError
        (these will have None as values).
        :return: iterable of LocalStorageRecords
        """
        if isinstance(host, str) and isinstance(key, str):
            if host not in self._host_lookup or key not in self._host_lookup[host]:
                if raise_on_no_result:
                    raise KeyError((host, key))
                else:
                    return []

            yield from (r for r in self._host_lookup[host][key] if include_deletions or not r.is_deleted)

        else:
            found_hosts = self._search_host(host)
            if raise_on_no_result and not found_hosts:
                raise KeyError((host, key))

            yielded = False
            for found_host in found_hosts:
                if isinstance(key, str):
                    matched_keys = [key]
                elif isinstance(key, re.Pattern):
                    matched_keys = [x for x in self._host_lookup[found_host].keys() if key.search(x)]
                elif isinstance(key, col_abc.Collection):
                    script_key_set = set(key)
                    matched_keys = list(self._host_lookup[found_host].keys() & script_key_set)
                elif isinstance(key, col_abc.Callable):
                    matched_keys = [x for x in self._host_lookup[found_host].keys() if key(x)]
                else:
                    raise TypeError(f"Unexpected type for script key: {type(key)} (expects: {KeySearch})")

                for matched_key in matched_keys:
                    for rec in self._host_lookup[found_host][matched_key]:
                        if include_deletions or not rec.is_deleted:
                            yielded = True
                            yield rec

            if not yielded and raise_on_no_result:
                raise KeyError((host, key))

    def iter_orphans(self) -> typing.Iterable[tuple[str, SessionStoreValue]]:
        """
        Returns records which have been orphaned from their host (domain name) where it cannot be recovered. The keys
            may be named uniquely enough that the host may be inferred.
        :return: yields tuples of (session key, SessionStoreValue)
        """
        yield from self._orphans

    def __getitem__(self, item: typing.Union[str, typing.Tuple[str, str]]) -> typing.Union[
            dict[str, tuple[SessionStoreValue, ...]], tuple[SessionStoreValue, ...]]:
        if item not in self:
            raise KeyError(item)

        if isinstance(item, str):
            return self.get_all_for_host(item)
        elif isinstance(item, tuple) and len(item) == 2:
            return self.get_session_storage_key(*item)
        else:
            raise TypeError("item must be a string or a tuple of (str, str)")

    def __iter__(self) -> typing.Iterable[str]:
        """
        iterates the hosts present
        """
        return self.iter_hosts()

    def close(self):
        self._ldb.close()

    def __enter__(self) -> "SessionStoreDb":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def main(args):
    ldb_in_dir = pathlib.Path(args[0])
    ssdb = SessionStoreDb(ldb_in_dir)

    print("Hosts in db:")
    for host in ssdb:
        print(host)


if __name__ == '__main__':
    main(sys.argv[1:])



"""
Copyright 2020, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import struct
import io
import typing
import enum

__version__ = "0.4"
__description__ = "Pure Python reimplementation of Google's Snappy decompression"
__contact__ = "Alex Caithness"


DEBUG = False
FRAME_MAGIC = bytes.fromhex("73 4E 61 50 70 59")


def make_crc_table(poly):
    table = []
    for i in range(256):
        crc = 0
        for _ in range(8):
            if (i ^ crc) & 1:
                crc = (crc >> 1) ^ poly
            else:
                crc >>= 1
            i >>= 1
        table.append(crc)
    return table


CRC_POLY = 0x82F63B78
CRC_QUICK_TABLE = tuple(make_crc_table(CRC_POLY))


def crc32c(data, xor_value=0xffffffff):
    value = 0xffffffff
    for b in data:
        value = CRC_QUICK_TABLE[(b ^ value) & 0xff] ^ (value >> 8)

    value ^= xor_value
    return value


# def log(msg):
#     if DEBUG:
#         print(msg)


class NoMoreData(Exception):
    ...


class ElementType(enum.IntEnum):
    """Run type in the compressed snappy data (literal data or offset to backreferenced data_"""
    Literal = 0
    CopyOneByte = 1
    CopyTwoByte = 2
    CopyFourByte = 3


def _read_le_varint(stream: typing.BinaryIO) -> typing.Optional[typing.Tuple[int, bytes]]:
    """Read varint from a stream.
    If the read is successful: returns a tuple of the (unsigned) value and the raw bytes making up that varint,
    otherwise returns None"""
    # this only outputs unsigned
    i = 0
    result = 0
    underlying_bytes = []
    while i < 10:  # 64 bit max possible?
        raw = stream.read(1)
        if len(raw) < 1:
            return None
        tmp, = raw
        underlying_bytes.append(tmp)
        result |= ((tmp & 0x7f) << (i * 7))
        if (tmp & 0x80) == 0:
            break
        i += 1
    return result, bytes(underlying_bytes)


def read_le_varint(stream: typing.BinaryIO) -> typing.Optional[int]:
    """Convenience version of _read_le_varint that only returns the value or None"""
    x = _read_le_varint(stream)
    if x is None:
        return None
    else:
        return x[0]


def read_uint16(stream: typing.BinaryIO) -> int:
    """Reads an Uint16 from stream"""
    return struct.unpack("<H", stream.read(2))[0]


def read_uint24(stream: typing.BinaryIO) -> int:
    """Reads an Uint24 from stream"""
    return struct.unpack("<I", stream.read(3) + b"\x00")[0]


def read_uint32(stream: typing.BinaryIO) -> int:
    """Reads an Uint32 from stream"""
    return struct.unpack("<I", stream.read(4))[0]


def read_byte(stream: typing.BinaryIO) -> typing.Optional[int]:
    """Reads a single byte from stream (or returns None if EOD is met)"""
    x = stream.read(1)
    if x:
        return x[0]

    return None


def decompress(data: typing.BinaryIO) -> bytes:
    """Decompresses the snappy compressed data stream"""
    uncompressed_length = read_le_varint(data)
    # log(f"Uncompressed length: {uncompressed_length}")

    out = io.BytesIO()

    while True:
        start_offset = data.tell()
        # log(f"Reading tag at offset {start_offset}")
        type_byte = read_byte(data)
        if type_byte is None:
            break

        # log(f"Type Byte is {type_byte:02x}")

        tag = type_byte & 0x03

        # log(f"Element Type is: {ElementType(tag)}")

        if tag == ElementType.Literal:
            if ((type_byte & 0xFC) >> 2) < 60:  # embedded in tag
                length = 1 + ((type_byte & 0xFC) >> 2)
                # log(f"Literal length is embedded in type byte and is {length}")
            elif ((type_byte & 0xFC) >> 2) == 60:  # 8 bit
                length = 1 + read_byte(data)
                # log(f"Literal length is 8bit and is {length}")
            elif ((type_byte & 0xFC) >> 2) == 61:  # 16 bit
                length = 1 + read_uint16(data)
                # log(f"Literal length is 16bit and is {length}")
            elif ((type_byte & 0xFC) >> 2) == 62:  # 16 bit
                length = 1 + read_uint24(data)
                # log(f"Literal length is 24bit and is {length}")
            elif ((type_byte & 0xFC) >> 2) == 63:  # 16 bit
                length = 1 + read_uint32(data)
                # log(f"Literal length is 32bit and is {length}")
            else:
                raise ValueError()  # cannot ever happen

            literal_data = data.read(length)
            if len(literal_data) < length:
                raise ValueError("Couldn't read enough literal data")

            out.write(literal_data)

        else:
            if tag == ElementType.CopyOneByte:
                length = ((type_byte & 0x1C) >> 2) + 4
                offset = ((type_byte & 0xE0) << 3) | read_byte(data)
            elif tag == ElementType.CopyTwoByte:
                length = 1 + ((type_byte & 0xFC) >> 2)
                offset = read_uint16(data)
            elif tag == ElementType.CopyFourByte:
                length = 1 + ((type_byte & 0xFC) >> 2)
                offset = read_uint32(data)
            else:
                raise ValueError()  # cannot ever happen

            if offset == 0:
                raise ValueError("Offset cannot be 0")

            actual_offset = out.tell() - offset
            # log(f"Current Outstream Length: {out.tell()}")
            # log(f"Backreference length: {length}")
            # log(f"Backreference relative offset: {offset}")
            # log(f"Backreference absolute offset: {actual_offset}")

            # have to read incrementally because you might have to read data that you've just written
            # for i in range(length):
            #     out.write(out.getbuffer()[actual_offset + i: actual_offset + i + 1].tobytes())
            buffer = out.getbuffer()[actual_offset: actual_offset + length].tobytes()
            if offset - length <= 0:
                # better safe than sorry, this way we're sure to extend it
                # as much as needed without doing some extra calculations
                buffer = (buffer * length)[:length]
            out.write(buffer)

    result = out.getvalue()
    if uncompressed_length != len(result):
        raise ValueError("Wrong data length in uncompressed data")
        # TODO: allow a partial / potentially bad result via a flag in the function call?

    return result


def check_masked_crc(crc, data, xor_value=0xffffffff):
    check = crc32c(data, xor_value=xor_value)

    check = ((check >> 15) | (check << 17)) & 0xffffffff  # rotate
    check += 0xa282ead8  # add constant
    check %= 0x100000000  # wraparound as an uint32

    return crc == check


def read_frame(frame_stream: typing.BinaryIO):
    frame_header = frame_stream.read(4)
    if not frame_header:
        raise NoMoreData()
    if len(frame_header) < 4:
        raise ValueError("Could not read entire frame header")

    frame_id = frame_header[0]
    frame_length, = struct.unpack("<I", frame_header[1:] + b"\x00")

    data = frame_stream.read(frame_length)
    if len(data) != frame_length:
        raise ValueError(f"Could not read all data; wanted: {frame_length}; got: {len(data)}")

    return frame_id, data


def decompress_framed(frame_stream: typing.BinaryIO, out_stream: typing.BinaryIO, *, mozilla_mode=False):
    """
    Decompresses a Snappy framed format stream into another stream.

    :param frame_stream: Stream containing the Snappy Framed data
    :param out_stream: Stream that the decompressed data will be written to.
    :param mozilla_mode: If True, use the (non-standard) checksum format used by Mozilla
    :return:
    """
    header_type, header_raw = read_frame(frame_stream)
    if header_type != 0xff or header_raw != FRAME_MAGIC:
        raise ValueError("Invalid magic")

    while True:
        frame_offset = frame_stream.tell()
        try:
            frame_type, frame_data = read_frame(frame_stream)
        except NoMoreData:
            break

        if frame_type == 0x00:  # compressed
            crc_raw = frame_data[0:4]
            with io.BytesIO(frame_data[4:]) as compressed:
                decompressed = decompress(compressed)
            stored_crc, = struct.unpack("<I", crc_raw)
            crc_match = check_masked_crc(stored_crc, decompressed, xor_value=0x0 if mozilla_mode else 0xffffffff)
            if not crc_match:
                raise ValueError(f"CRC mismatch in frame starting at {frame_offset}")

            out_stream.write(decompressed)
        elif frame_type == 0x01:  # decompressed
            crc_raw = frame_data[0:4]
            stored_crc, = struct.unpack("<I", crc_raw)
            crc_match = check_masked_crc(stored_crc, frame_data[4:], xor_value=0x0 if mozilla_mode else 0xffffffff)
            if not crc_match:
                raise ValueError(f"CRC mismatch in frame starting at {frame_offset}")
            out_stream.write(frame_data[4:])
        elif frame_type == 0xfe:  # padding
            pass
        elif 0x02 <= frame_type <= 0x7f:  # reserved, unskippable
            raise ValueError("Reserved unskippable data")
        elif 0x80 <= frame_type <= 0xfe:  # reserved, skippable
            pass
        else:
            raise ValueError("unexpected frame")


def _main(in_path, out_path):
    import pathlib

    in_path = pathlib.Path(in_path)
    out_path = pathlib.Path(out_path)
    with in_path.open("rb") as f:
        with out_path.open("wb") as out:
            decompress_framed(f, out)


if __name__ == "__main__":
    _main(sys.argv[1], sys.argv[2])



"""
Copyright 2022, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import base64
import json
import pathlib
import re
import sys
import os
import typing
import abc
import sqlite3
import datetime
import win32crypt
import Crypto.Cipher.AES
from ccl_chromium_reader import ccl_chromium_localstorage
from ccl_chromium_reader import ccl_chromium_sessionstorage
from ccl_chromium_reader import ccl_chromium_indexeddb
from ccl_chromium_reader import ccl_chromium_filesystem
from ccl_chromium_reader import ccl_shared_proto_db_downloads
from ccl_chromium_reader import ccl_chromium_cache
from ccl_chromium_reader import ccl_chromium_notifications
from ccl_chromium_reader import ccl_chromium_snss2

sys.stdout.reconfigure(encoding="utf-8")

__version__ = "0.3"
__description__ = "Audits multiple Chromium data stores"
__contact__ = "Alex Caithness"


WINDOWS = False

CHROME_EPOCH = datetime.datetime(1601, 1, 1, 0, 0, 0)


def chrome_time(microseconds: typing.Optional[int], allow_none=True):
    if microseconds is not None:
        return CHROME_EPOCH + datetime.timedelta(microseconds=microseconds)
    elif allow_none:
        return None
    raise ValueError("microseconds cannot be None")


class AbstractAuditor(abc.ABC):
    def __init__(self, name: str):
        self.name = name

    @property
    @abc.abstractmethod
    def headers(self) -> tuple[str, ...]:
        raise NotImplementedError

    @abc.abstractmethod
    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        raise NotImplementedError


class BookmarksAuditor(AbstractAuditor):
    @property
    def headers(self) -> tuple[str, ...]:
        return "id", "guid", "path", "url", "added time",

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        bookmarks_path = pathlib.Path(profile_root, "Bookmarks")
        if not bookmarks_path.exists():
            return

        with bookmarks_path.open("rt", encoding="utf-8") as f:
            bookmarks = json.load(f)

        def walk_bookmarks(obj: dict, parts: list):
            print(obj)
            for inner in obj["children"]:
                if inner["type"] == "folder":
                    yield from walk_bookmarks(inner, parts + [inner["name"]])
                elif inner["type"] == "url":
                    if domain_re.search(inner["url"]) is not None:
                        yield (
                            inner["id"],
                            inner["guid"],
                            "/".join(parts + [inner["name"]]),
                            inner["url"],
                            chrome_time(int(inner["date_added"]))
                        )
                else:
                    raise ValueError("unexpected bookmark type")

        for key, root in bookmarks["roots"].items():
            if key == "sync_transaction_version":
                continue

            yield from walk_bookmarks(root, [key])

    def __init__(self):
        super().__init__("Bookmarks")


class HistoryAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("History")

    @property
    def headers(self) -> tuple[str, ...]:
        return "id", "url", "title", "timestamp"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        profile_folder = pathlib.Path(profile_root)
        history_path = profile_folder / "History"
        if not history_path.exists():
            return
        conn = sqlite3.connect(history_path)
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute("""
        SELECT 
          visits.id AS "id",
          urls.url AS "url",
          urls.title AS "title",
          visits.visit_time AS "timestamp", 
          visits.visit_duration AS "duration",
          visits.from_visit AS "from_visit",
          visits.transition AS "transition"
        FROM visits
          INNER JOIN urls ON visits.url = urls.id; 
        """)

        for row in cur:
            if domain_re.search(row["url"]) is not None:
                yield (
                    row["id"],
                    row["url"],
                    row["title"],
                    chrome_time(row["timestamp"]),
                )

        conn.close()


class DownloadsHistoryAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Downloads (History)")

    @property
    def headers(self) -> tuple[str, ...]:
        return (
            "id", "guid", "tab url", "tab referrer url", "target path", "total bytes",
            "mime-type", "original mime-type", "start time"
        )

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        profile_folder = pathlib.Path(profile_root)

        history_path = profile_folder / "History"
        if not history_path.exists():
            return

        conn = sqlite3.connect(history_path)
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute("""
                SELECT 
                  downloads.id,
                  downloads.guid,
                  downloads.tab_url,
                  downloads.tab_referrer_url,
                  downloads.target_path,
                  downloads.total_bytes,
                  downloads.mime_type,
                  downloads.original_mime_type,
                  downloads.start_time
                FROM downloads
                """)

        for row in cur:
            if domain_re.search(row["tab_url"]) is not None or domain_re.search(row["tab_referrer_url"]) is not None:
                yield (
                    row["id"],
                    row["guid"],
                    row["tab_url"],
                    row["tab_referrer_url"],
                    row["target_path"],
                    row["total_bytes"],
                    row["mime_type"],
                    row["original_mime_type"],
                    chrome_time(row["start_time"])
                )
        conn.close()


class DownloadsSharedProtoDb(AbstractAuditor):
    def __init__(self):
        super().__init__("Downloads (Shard Proto Db)")

    @property
    def headers(self) -> tuple[str, ...]:
        return (
            "id", "guid", "tab url", "tab referrer url", "target path", "total bytes",
            "mime-type", "original mime-type", "start time"
        )

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        shared_proto_db_folder = pathlib.Path(profile_root) / "shared_proto_db"
        if not shared_proto_db_folder.exists():
            return
        for download in ccl_shared_proto_db_downloads.read_downloads(shared_proto_db_folder):
            if (domain_re.search(download.tab_url or "") is not None or
                    domain_re.search(download.tab_referrer_url or "") is not None):
                yield (
                    download.level_db_seq_no,
                    download.guid,
                    download.tab_url,
                    download.tab_referrer_url,
                    download.target_path,
                    download.total_bytes,
                    download.mime_type,
                    download.original_mime_type,
                    download.start_time
                )


class FaviconAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Favicons")

    @property
    def headers(self) -> tuple[str, ...]:
        return "id", "page_url", "favicon url"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        favicons_path = pathlib.Path(profile_root) / "Favicons"
        if not favicons_path.exists():
            return

        conn = sqlite3.connect(favicons_path)
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()

        cur.execute("""
        SELECT 
          icon_mapping.id, 
          icon_mapping.page_url,
          favicons.url
        FROM icon_mapping
        LEFT JOIN favicons ON icon_mapping.icon_id = favicons.id
        """)

        for row in cur:
            if domain_re.search(row["page_url"]) is not None or domain_re.search(row["url"]) is not None:
                yield (
                    row["id"],
                    row["page_url"],
                    row["url"]
                )

        conn.close()


class CacheAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Cache")

    @property
    def headers(self) -> tuple[str, ...]:
        return (
            "key", "request time", "response time", "data stream file type", "data stream file selector",
            "block number", "block count", "data stream external file number")

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern,
              override_path: typing.Optional[typing.Union[os.PathLike, str]] = None) -> typing.Iterator[tuple]:
        cache_folder = override_path or (pathlib.Path(profile_root) / "Cache" / "Cache_Data")
        cache_type = ccl_chromium_cache.guess_cache_class(cache_folder)
        if cache_type == ccl_chromium_cache.ChromiumBlockFileCache:
            with ccl_chromium_cache.ChromiumBlockFileCache(cache_folder) as cache:
                for key, es in cache.items():
                    metas = cache.get_metadata(key)
                    for meta in metas:
                        if domain_re.search(key) is not None:
                            yield (
                                key,
                                meta.request_time,
                                meta.response_time,
                                es.data_addrs[1].file_type,
                                es.data_addrs[1].file_selector,
                                es.data_addrs[1].block_number,
                                es.data_addrs[1].contiguous_blocks,
                                es.data_addrs[1].external_file_number
                            )
        else:
            with ccl_chromium_cache.ChromiumSimpleFileCache(cache_folder) as cache:
                for key in cache.keys():
                    if domain_re.search(key) is not None:
                        metas = zip(cache.get_metadata(key), cache.get_file_for_key(key))
                        for meta, file_name in metas:
                            yield (
                                key,
                                meta.request_time,
                                meta.response_time,
                                file_name
                            )


class CookieAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Cookies")

    @property
    def headers(self) -> tuple[str, ...]:
        return "ID", "Host", "Name", "Value", "Creation Time"

    @staticmethod
    def decrypt_windows_cookie(encryption_key: bytes, cipher_text: bytes):
        cipher = Crypto.Cipher.AES.new(encryption_key, nonce=cipher_text[3:3+12], mode=Crypto.Cipher.AES.MODE_GCM)
        value = cipher.decrypt_and_verify(cipher_text[3+12:-16], cipher_text[-16:])
        return value

    @staticmethod
    def nop(*args):
        return None

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        cookie_path = pathlib.Path(profile_root) / "Cookies"  # old location on windows, still here on some platforms
        if not cookie_path.exists():
            cookie_path = pathlib.Path(profile_root) / "Network" / "Cookies"
        if not cookie_path.exists():
            return

        local_state_path = pathlib.Path(profile_root).parent / "Local State"

        decrypter = self.nop

        if local_state_path.exists():
            if WINDOWS:
                with local_state_path.open("rt", encoding="utf-8") as f:
                    local_state = json.load(f)
                encryption_key = base64.b64decode(local_state['os_crypt']['encrypted_key'])
                encryption_key = encryption_key[5:]
                encryption_key = win32crypt.CryptUnprotectData(encryption_key, None, None, None, 0)[1]
                decrypter = lambda x: self.decrypt_windows_cookie(encryption_key, x)
        else:
            print("Cannot get Local State file to decode cookie values")

        conn = sqlite3.connect(cookie_path)
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute("""SELECT 
                      rowid,
                      cookies.host_key,
                      cookies.name,
                      cookies.value,
                      cookies.encrypted_value,
                      cookies.creation_utc
                    FROM cookies""")

        for row in cur:
            if domain_re.search(row["host_key"]):
                value = row["value"] if row["value"] is not None else decrypter(row["encrypted_value"]).decode("utf-8")
                rowid = row["rowid"] if "rowid" in row else row["creation_utc"]
                yield (
                    rowid,
                    row["host_key"],
                    row["name"],
                    value,
                    chrome_time(row["creation_utc"])
                )

        conn.close()


class LocalStorageAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Local Storage")

    @property
    def headers(self) -> tuple[str, ...]:
        return "id", "storage key", "script key", "value"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        ldb_path = pathlib.Path(profile_root) / "Local Storage" / "leveldb"
        if not ldb_path.exists():
            return

        local_storage =  ccl_chromium_localstorage.LocalStoreDb(ldb_path)
        for storage_key in local_storage.iter_storage_keys():
            if domain_re.search(storage_key) is not None:
                for rec in local_storage.iter_records_for_storage_key(storage_key):
                    yield (
                        rec.leveldb_seq_number,
                        rec.storage_key,
                        rec.script_key,
                        rec.value
                    )

        local_storage.close()


class SessionStorageAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Session Storage")

    @property
    def headers(self) -> tuple[str, ...]:
        return "id", "host", "key", "value"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        ldb_path = pathlib.Path(profile_root) / "Session Storage"
        if not ldb_path.exists():
            return

        session_storage = ccl_chromium_sessionstorage.SessionStoreDb(ldb_path)
        value: typing.Optional[ccl_chromium_sessionstorage.SessionStoreValue] = None
        for host in session_storage.iter_hosts():
            if domain_re.search(host) is not None:
                for ss_key, values in session_storage.get_all_for_host(host).items():
                    for value in values:
                        yield (
                            value.leveldb_sequence_number,
                            host,
                            ss_key,
                            value.value
                        )

        session_storage.close()


class IndexedDbAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("IndexedDb")

    @property
    def headers(self) -> tuple[str, ...]:
        return "id", "origin", "database", "object store", "key", "value"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        idb_root = pathlib.Path(profile_root) / "IndexedDB"
        if not idb_root.exists():
            return

        def bad_deserializer_data_handler(key: ccl_chromium_indexeddb.IdbKey, buffer: bytes):
            print(f"Error reading IndexedDb record {key}", file=sys.stderr)

        for ldb_folder in idb_root.glob("*.leveldb"):
            if domain_re.search(ldb_folder.stem) is not None:
                idb = ccl_chromium_indexeddb.WrappedIndexDB(ldb_folder)
                for database_id in idb.database_ids:
                    database = idb[database_id.dbid_no]
                    for obj_store_name in database.object_store_names:
                        obj_store = database.get_object_store_by_name(obj_store_name)
                        for rec in obj_store.iterate_records(
                                bad_deserializer_data_handler=bad_deserializer_data_handler):
                            yield (
                                rec.sequence_number,
                                database.origin,
                                database.name,
                                obj_store.name,
                                rec.key,
                                rec.value
                            )


class FileSystemAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("FileSystem")

    @property
    def headers(self) -> tuple[str, ...]:
        return "Folder ID", "Sequence Number", "Name", "Data Path", "Storage Type"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        file_system_path = pathlib.Path(profile_root) / "File System"
        if not file_system_path.exists():
            return

        file_system = ccl_chromium_filesystem.FileSystem(file_system_path)
        for origin in file_system.get_origins():
            if domain_re.search(origin):
                for folder in file_system.get_folders_for_origin(origin):
                    origin_storage = file_system.get_storage_for_folder(folder)
                    for file_id, file_info in origin_storage.get_file_listing():
                        yield (
                            file_info.folder_id,
                            file_info.seq_no,
                            file_info.name,
                            file_info.data_path,
                            "Persistent" if file_info.is_persistent else "Temporary"
                        )


class NotificationAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Notifications")

    @property
    def headers(self) -> tuple[str, ...]:
        return "ID", "Origin", "Title", "Body", "Data", "Timestamp"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        notification_path = pathlib.Path(profile_root) / "Platform Notifications"

        if not notification_path.exists():
            return

        with ccl_chromium_notifications.NotificationReader(notification_path) as reader:
            for notification in reader.read_notifications():
                if domain_re.search(notification.origin) is not None:
                    yield (
                        notification.level_db_info.seq_no,
                        notification.origin,
                        json.dumps(notification.title),
                        json.dumps(notification.body),
                        json.dumps(notification.data),
                        notification.timestamp
                    )


class LoginAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Logins")

    @property
    def headers(self) -> tuple[str, ...]:
        return "ID", "database", "origin url", "action url", "username_value"

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        for db_name in ("Login Data", "Login Data For Account"):
            db_path = pathlib.Path(profile_root) / db_name
            if not db_path.exists():
                continue
            conn = sqlite3.connect(db_path)
            conn.row_factory = sqlite3.Row
            cur = conn.cursor()
            cur.execute("""SELECT 
                          rowid,
                          logins.origin_url,
                          logins.action_url,
                          logins.username_value
                        FROM logins""")

            for row in cur:
                if domain_re.search(row["origin_url"]) is not None or domain_re.search(row["action_url"]):
                    yield (
                        row["id"],
                        db_name,
                        row["origin_url"],
                        row["action_url"],
                        row["username_value"]
                    )
            conn.close()


class SnssAuditor(AbstractAuditor):
    def __init__(self):
        super().__init__("Snss")

    @property
    def headers(self) -> tuple[str, ...]:
        return "file", "offset", "index", "timestamp", "title", "url", "original_request_url", "referrer_url",

    def audit(self, profile_root: typing.Union[os.PathLike, str], domain_re: re.Pattern) -> typing.Iterator[tuple]:
        session_folder = pathlib.Path(profile_root) / "Sessions"
        if not session_folder.exists():
            return
        for snss_file in session_folder.iterdir():
            if not snss_file.is_file():
                continue
            if not (snss_file.name.startswith("Session_") or snss_file.name.startswith("Tabs_")):
                continue

            with snss_file.open("rb") as f:
                snss = ccl_chromium_snss2.SnssFile(
                    ccl_chromium_snss2.SnssFileType.Session if snss_file.name.startswith("Session_")
                    else ccl_chromium_snss2.SnssFileType.Tab, f)
                for navigation_entry in snss.iter_session_commands():
                    if not isinstance(navigation_entry, ccl_chromium_snss2.NavigationEntry):
                        continue  # TODO: There may well be other useful session commands to look into later

                    # TODO: add PageState stuff once it's in place in ccl_chromium_snss2
                    yield (
                        snss_file.name,
                        navigation_entry.offset,
                        navigation_entry.index,
                        navigation_entry.timestamp,
                        navigation_entry.title,
                        navigation_entry.url,
                        navigation_entry.original_request_url,
                        navigation_entry.referrer_url,
                    )


AUDITORS: typing.Collection[AbstractAuditor] = (
    BookmarksAuditor(),
    HistoryAuditor(),
    DownloadsHistoryAuditor(),
    DownloadsSharedProtoDb(),
    FaviconAuditor(),
    CacheAuditor(),
    CookieAuditor(),
    LocalStorageAuditor(),
    SessionStorageAuditor(),
    IndexedDbAuditor(),
    FileSystemAuditor(),
    NotificationAuditor(),
    LoginAuditor(),
    SnssAuditor()
)


def main(args):
    profile_folder = pathlib.Path(args[0])
    domain_re = re.compile(args[1])
    cache_folder = None if len(args) < 3 else args[2]

    for auditor in AUDITORS:
        print("-" * 72)
        print(auditor.name)
        print("-" * 72)
        print("\t".join(auditor.headers))

        if isinstance(auditor, CacheAuditor) and cache_folder is not None:
            results = auditor.audit(profile_folder, domain_re, cache_folder)
        else:
            results = auditor.audit(profile_folder, domain_re)

        for result in results:
            print("\t".join(str(x) for x in result))


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(f"{pathlib.Path(sys.argv[0]).name} <chrome profile folder> <pattern for url matching> [cache folder (for mobile)]")
        exit(1)
    main(sys.argv[1:])



from __future__ import annotations

import pytest

from linear_mcp_fast.reader import CachedData, LinearLocalReader


def _build_cache() -> CachedData:
    cache = CachedData()
    cache.users = {
        "U1": {
            "id": "U1",
            "email": "target@example.com",
            "organizationId": "ORG1",
            "userAccountId": "ACC1",
            "name": "Target",
            "displayName": "target",
        },
        "U2": {
            "id": "U2",
            "email": "coworker@example.com",
            "organizationId": "ORG1",
            "userAccountId": "ACC2",
            "name": "Coworker",
            "displayName": "coworker",
        },
        "U3": {
            "id": "U3",
            "email": "other@example.com",
            "organizationId": "ORG2",
            "userAccountId": "ACC3",
            "name": "Other",
            "displayName": "other",
        },
    }
    cache.teams = {
        "T1": {"id": "T1", "key": "ORG1", "name": "Org1 Team", "organizationId": "ORG1"},
        "T2": {"id": "T2", "key": "ORG2", "name": "Org2 Team", "organizationId": "ORG2"},
    }
    cache.states = {
        "S1": {"id": "S1", "teamId": "T1", "type": "started", "name": "In Progress"},
        "S2": {"id": "S2", "teamId": "T2", "type": "started", "name": "In Progress"},
    }
    cache.issues = {
        "I1": {"id": "I1", "teamId": "T1", "stateId": "S1", "projectId": "P1"},
        "I2": {"id": "I2", "teamId": "T2", "stateId": "S2", "projectId": "P2"},
    }
    cache.issue_content = {"I1": "org1", "I2": "org2"}
    cache.comments = {
        "C1": {"id": "C1", "issueId": "I1"},
        "C2": {"id": "C2", "issueId": "I2"},
    }
    cache.comments_by_issue = {"I1": ["C1"], "I2": ["C2"]}
    cache.projects = {
        "P1": {"id": "P1", "teamIds": ["T1"], "statusId": "PS1", "memberIds": ["U2"]},
        "P2": {"id": "P2", "teamIds": ["T2"], "statusId": "PS2", "memberIds": ["U3"]},
    }
    cache.labels = {
        "L1": {"id": "L1", "teamId": "T1"},
        "L2": {"id": "L2", "teamId": "T2"},
    }
    cache.initiatives = {
        "N1": {"id": "N1", "teamIds": ["T1"], "ownerId": "U1"},
        "N2": {"id": "N2", "teamIds": ["T2"], "ownerId": "U3"},
    }
    cache.cycles = {
        "CY1": {"id": "CY1", "teamId": "T1"},
        "CY2": {"id": "CY2", "teamId": "T2"},
    }
    cache.documents = {
        "D1": {"id": "D1", "projectId": "P1", "creatorId": "U2"},
        "D2": {"id": "D2", "projectId": "P2", "creatorId": "U3"},
    }
    cache.milestones = {
        "M1": {"id": "M1", "projectId": "P1"},
        "M2": {"id": "M2", "projectId": "P2"},
    }
    cache.project_updates = {
        "UP1": {"id": "UP1", "projectId": "P1"},
        "UP2": {"id": "UP2", "projectId": "P2"},
    }
    cache.project_statuses = {
        "PS1": {"id": "PS1", "name": "On Track"},
        "PS2": {"id": "PS2", "name": "At Risk"},
    }
    return cache


def test_apply_account_scope_filters_by_email(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv("LINEAR_FAST_ACCOUNT_EMAILS", "target@example.com")
    monkeypatch.delenv("LINEAR_FAST_USER_ACCOUNT_IDS", raising=False)
    reader = LinearLocalReader()
    cache = _build_cache()

    reader._apply_account_scope(cache)

    assert set(cache.users.keys()) == {"U1", "U2"}
    assert set(cache.teams.keys()) == {"T1"}
    assert set(cache.states.keys()) == {"S1"}
    assert set(cache.issues.keys()) == {"I1"}
    assert set(cache.comments.keys()) == {"C1"}
    assert cache.comments_by_issue == {"I1": ["C1"]}
    assert set(cache.projects.keys()) == {"P1"}
    assert set(cache.labels.keys()) == {"L1"}
    assert set(cache.initiatives.keys()) == {"N1"}
    assert set(cache.cycles.keys()) == {"CY1"}
    assert set(cache.documents.keys()) == {"D1"}
    assert set(cache.milestones.keys()) == {"M1"}
    assert set(cache.project_updates.keys()) == {"UP1"}
    assert set(cache.project_statuses.keys()) == {"PS1"}


def test_apply_account_scope_raises_when_email_not_found(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv("LINEAR_FAST_ACCOUNT_EMAILS", "missing@example.com")
    monkeypatch.delenv("LINEAR_FAST_USER_ACCOUNT_IDS", raising=False)
    reader = LinearLocalReader()
    cache = _build_cache()

    with pytest.raises(ValueError):
        reader._apply_account_scope(cache)



from __future__ import annotations

import asyncio

import pytest

from linear_mcp_fast.official_session import (
    DEFAULT_OFFICIAL_MCP_URL,
    OfficialToolError,
    OfficialMcpSessionManager,
)


@pytest.fixture(autouse=True)
def _clear_official_env(monkeypatch: pytest.MonkeyPatch):
    keys = [
        "LINEAR_OFFICIAL_MCP_TRANSPORT",
        "LINEAR_OFFICIAL_MCP_COMMAND",
        "LINEAR_OFFICIAL_MCP_ARGS",
        "LINEAR_OFFICIAL_MCP_ENV",
        "LINEAR_OFFICIAL_MCP_CWD",
        "LINEAR_OFFICIAL_MCP_URL",
        "LINEAR_OFFICIAL_MCP_HEADERS",
    ]
    for key in keys:
        monkeypatch.delenv(key, raising=False)


def test_default_transport_uses_stdio(monkeypatch: pytest.MonkeyPatch):
    manager = OfficialMcpSessionManager()

    health = manager.get_health()
    assert health["transport"] == "stdio"
    assert health["command"] == "npx"
    assert health["args"] == ["-y", "mcp-remote", DEFAULT_OFFICIAL_MCP_URL]


def test_stdio_args_support_json_array(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(
        "LINEAR_OFFICIAL_MCP_ARGS",
        '["-y", "mcp-remote", "https://example.com/mcp", "--foo", "bar"]',
    )
    manager = OfficialMcpSessionManager()

    health = manager.get_health()
    assert health["args"] == [
        "-y",
        "mcp-remote",
        "https://example.com/mcp",
        "--foo",
        "bar",
    ]


def test_stdio_args_support_shell_style_string(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(
        "LINEAR_OFFICIAL_MCP_ARGS",
        "-y mcp-remote https://example.com/mcp --name 'My Client'",
    )
    manager = OfficialMcpSessionManager()

    health = manager.get_health()
    assert health["args"] == [
        "-y",
        "mcp-remote",
        "https://example.com/mcp",
        "--name",
        "My Client",
    ]


def test_stdio_args_invalid_shell_string_falls_back_to_default(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv("LINEAR_OFFICIAL_MCP_ARGS", "-y mcp-remote 'unterminated")
    manager = OfficialMcpSessionManager()

    health = manager.get_health()
    assert health["args"] == ["-y", "mcp-remote", DEFAULT_OFFICIAL_MCP_URL]


def test_http_transport_keeps_legacy_header_mode(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv("LINEAR_OFFICIAL_MCP_TRANSPORT", "http")
    monkeypatch.setenv("LINEAR_OFFICIAL_MCP_HEADERS", '{"Authorization": "Bearer X"}')

    manager = OfficialMcpSessionManager()
    health = manager.get_health()

    assert health["transport"] == "http"
    assert health["url"] == DEFAULT_OFFICIAL_MCP_URL
    assert health["hasHeaders"] is True


def test_invalid_transport_raises(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv("LINEAR_OFFICIAL_MCP_TRANSPORT", "invalid")
    with pytest.raises(ValueError):
        OfficialMcpSessionManager()


def test_module_alias_runs_main():
    from oh_my_linearmcp import main as alias_main
    from linear_mcp_fast import main as base_main

    assert alias_main is base_main


class _FakeText:
    type = "text"

    def __init__(self, text: str):
        self.text = text


class _FakeResult:
    def __init__(self, *, is_error: bool = False, text: str = "", tools: list[object] | None = None):
        self.isError = is_error
        self.content = [_FakeText(text)] if text else []
        self.tools = tools or []


class _FakeTool:
    def __init__(self, name: str):
        self.name = name


def _sync_submit(value):
    if asyncio.iscoroutine(value):
        return asyncio.run(value)
    return value


def test_call_tool_preserves_official_tool_error(monkeypatch: pytest.MonkeyPatch):
    manager = OfficialMcpSessionManager()

    class _FakeSession:
        def call_tool(self, name: str, arguments=None):
            return _FakeResult(is_error=True, text="official said no")

    manager._session = _FakeSession()
    monkeypatch.setattr(manager, "_ensure_connected", lambda: None)
    monkeypatch.setattr(manager, "_submit", _sync_submit)

    with pytest.raises(OfficialToolError) as exc_info:
        manager.call_tool("list_issues", {})

    assert exc_info.value.code == "official_tool_error"
    assert "official said no" in exc_info.value.message


def test_list_tools_retries_once(monkeypatch: pytest.MonkeyPatch):
    manager = OfficialMcpSessionManager()
    calls = {"count": 0}

    class _FakeSession:
        def list_tools(self):
            calls["count"] += 1
            if calls["count"] == 1:
                raise RuntimeError("transient failure")
            return _FakeResult(tools=[_FakeTool("create_issue"), _FakeTool("list_issues")])

    manager._session = _FakeSession()
    monkeypatch.setattr(manager, "_ensure_connected", lambda: None)
    monkeypatch.setattr(manager, "_submit", _sync_submit)
    async def _noop_disconnect():
        return None
    monkeypatch.setattr(manager, "_disconnect_async", _noop_disconnect)

    tools = manager.list_tools()

    assert tools == ["create_issue", "list_issues"]
    assert calls["count"] == 2



"""
Official Linear MCP client session manager.

Maintains a long-lived MCP client session with reconnect and single-retry
semantics for transient failures.

Default transport uses the official `mcp-remote` stdio bridge so existing OAuth
flows (as used by Claude MCP config) are reused without custom token plumbing.
"""

from __future__ import annotations

import asyncio
import concurrent.futures
import json
import logging
import os
import shlex
import threading
import time
from datetime import timedelta
from typing import Any

from mcp.client.session import ClientSession
from mcp.client.stdio import StdioServerParameters, stdio_client
from mcp.client.streamable_http import streamablehttp_client

logger = logging.getLogger(__name__)

DEFAULT_OFFICIAL_MCP_URL = "https://mcp.linear.app/mcp"
DEFAULT_TRANSPORT = "stdio"
DEFAULT_STDIO_COMMAND = "npx"
DEFAULT_STDIO_ARGS_PREFIX = ["-y", "mcp-remote"]


class OfficialToolError(RuntimeError):
    """Raised when the official MCP call fails."""

    def __init__(self, code: str, message: str):
        super().__init__(message)
        self.code = code
        self.message = message


class OfficialMcpSessionManager:
    """Thread-safe synchronous wrapper around async MCP client session."""

    def __init__(
        self,
        transport: str | None = None,
        command: str | None = None,
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
        cwd: str | None = None,
        url: str | None = None,
        headers: dict[str, str] | None = None,
        timeout_seconds: float = 30.0,
        sse_read_timeout_seconds: float = 300.0,
        read_timeout_seconds: float = 30.0,
    ):
        self._transport = (transport or os.getenv("LINEAR_OFFICIAL_MCP_TRANSPORT", DEFAULT_TRANSPORT)).lower()
        self._url = url or os.getenv("LINEAR_OFFICIAL_MCP_URL", DEFAULT_OFFICIAL_MCP_URL)
        self._headers = headers or self._parse_headers_from_env()
        self._command = command or os.getenv("LINEAR_OFFICIAL_MCP_COMMAND", DEFAULT_STDIO_COMMAND)
        self._args = args or self._parse_stdio_args_from_env(default_url=self._url)
        self._env = env or self._parse_stdio_env_from_env()
        self._cwd = cwd or os.getenv("LINEAR_OFFICIAL_MCP_CWD")
        self._timeout_seconds = timeout_seconds
        self._sse_read_timeout_seconds = sse_read_timeout_seconds
        self._read_timeout_seconds = read_timeout_seconds

        if self._transport not in {"stdio", "http"}:
            raise ValueError("LINEAR_OFFICIAL_MCP_TRANSPORT must be one of: stdio, http")

        self._loop: asyncio.AbstractEventLoop | None = None
        self._thread: threading.Thread | None = None
        self._lock = threading.RLock()

        self._transport_cm: Any = None
        self._session_cm: Any = None
        self._session: ClientSession | None = None

        self._failure_count = 0
        self._last_error: str | None = None
        self._last_failure_at: float | None = None
        self._last_connected_at: float | None = None

    @staticmethod
    def _parse_headers_from_env() -> dict[str, str] | None:
        raw = os.getenv("LINEAR_OFFICIAL_MCP_HEADERS")
        if not raw:
            return None
        try:
            parsed = json.loads(raw)
            if isinstance(parsed, dict):
                return {str(k): str(v) for k, v in parsed.items()}
        except Exception:
            logger.warning("Ignoring invalid LINEAR_OFFICIAL_MCP_HEADERS value")
        return None

    @staticmethod
    def _parse_stdio_env_from_env() -> dict[str, str] | None:
        raw = os.getenv("LINEAR_OFFICIAL_MCP_ENV")
        if not raw:
            return None
        try:
            parsed = json.loads(raw)
            if isinstance(parsed, dict):
                return {str(k): str(v) for k, v in parsed.items()}
        except Exception:
            logger.warning("Ignoring invalid LINEAR_OFFICIAL_MCP_ENV value")
        return None

    @staticmethod
    def _parse_stdio_args_from_env(default_url: str) -> list[str]:
        raw = os.getenv("LINEAR_OFFICIAL_MCP_ARGS")
        if not raw:
            return [*DEFAULT_STDIO_ARGS_PREFIX, default_url]

        # Prefer JSON array for exact argument boundaries.
        try:
            parsed = json.loads(raw)
            if isinstance(parsed, list):
                return [str(item) for item in parsed]
        except Exception:
            pass

        # Fallback to shell-like parsing.
        try:
            return shlex.split(raw)
        except ValueError:
            logger.warning("Ignoring invalid LINEAR_OFFICIAL_MCP_ARGS value; using default args")
            return [*DEFAULT_STDIO_ARGS_PREFIX, default_url]

    def _ensure_loop(self) -> None:
        if self._loop and self._thread and self._thread.is_alive():
            return

        loop = asyncio.new_event_loop()

        def _run_loop() -> None:
            asyncio.set_event_loop(loop)
            loop.run_forever()

        thread = threading.Thread(target=_run_loop, daemon=True, name="linear-official-mcp")
        thread.start()

        self._loop = loop
        self._thread = thread

    def _submit(self, coro: Any) -> Any:
        if not self._loop:
            raise RuntimeError("event loop not initialized")
        future = asyncio.run_coroutine_threadsafe(coro, self._loop)
        try:
            return future.result(timeout=self._read_timeout_seconds + 10)
        except concurrent.futures.TimeoutError:
            future.cancel()
            raise

    async def _connect_async(self) -> None:
        if self._session is not None:
            return

        if self._transport == "stdio":
            params = StdioServerParameters(
                command=self._command,
                args=self._args,
                env=self._env,
                cwd=self._cwd,
            )
            self._transport_cm = stdio_client(params)
        else:
            self._transport_cm = streamablehttp_client(
                self._url,
                headers=self._headers,
                timeout=self._timeout_seconds,
                sse_read_timeout=self._sse_read_timeout_seconds,
                terminate_on_close=False,
            )
        transport_streams = await self._transport_cm.__aenter__()
        if len(transport_streams) == 3:
            read_stream, write_stream, _ = transport_streams
        elif len(transport_streams) == 2:
            read_stream, write_stream = transport_streams
        else:
            raise RuntimeError("official MCP transport returned unexpected stream tuple")

        self._session_cm = ClientSession(
            read_stream,
            write_stream,
            read_timeout_seconds=timedelta(seconds=self._read_timeout_seconds),
        )
        self._session = await self._session_cm.__aenter__()
        await self._session.initialize()
        self._last_connected_at = time.time()

    async def _disconnect_async(self) -> None:
        if self._session_cm is not None:
            try:
                await self._session_cm.__aexit__(None, None, None)
            except Exception as exc:
                self._log_cleanup_exception("Official MCP session cleanup failed", exc)
        if self._transport_cm is not None:
            try:
                await self._transport_cm.__aexit__(None, None, None)
            except Exception as exc:
                self._log_cleanup_exception("Official MCP transport cleanup failed", exc)

        self._session = None
        self._session_cm = None
        self._transport_cm = None

    def _ensure_connected(self) -> None:
        self._ensure_loop()
        self._submit(self._connect_async())

    @staticmethod
    def _log_cleanup_exception(prefix: str, exc: Exception) -> None:
        message = str(exc)
        if "Attempted to exit cancel scope in a different task" in message:
            logger.debug("%s: %s", prefix, exc)
            return
        logger.warning("%s: %s", prefix, exc)

    def _normalize_result(self, result: Any) -> Any:
        if getattr(result, "isError", False):
            text = self._extract_text(result)
            raise OfficialToolError(
                "official_tool_error", text or "official MCP returned an error"
            )

        structured = getattr(result, "structuredContent", None)
        if structured is not None:
            return structured

        text = self._extract_text(result)
        if text:
            try:
                return json.loads(text)
            except Exception:
                return {"text": text}

        if hasattr(result, "model_dump"):
            return result.model_dump()
        return result

    @staticmethod
    def _extract_text(result: Any) -> str:
        content = getattr(result, "content", None) or []
        texts: list[str] = []
        for block in content:
            if getattr(block, "type", None) == "text":
                text = getattr(block, "text", "")
                if text:
                    texts.append(text)
        return "\n".join(texts).strip()

    def _record_failure(self, exc: Exception) -> None:
        self._failure_count += 1
        self._last_failure_at = time.time()
        self._last_error = f"{exc.__class__.__name__}: {exc}"
        logger.warning("Official MCP call failed (%s): %s", exc.__class__.__name__, exc)

    def _record_success(self) -> None:
        self._failure_count = 0
        self._last_error = None

    def call_tool(self, name: str, arguments: dict[str, Any] | None = None) -> Any:
        args = arguments or {}
        with self._lock:
            for attempt in range(2):
                try:
                    self._ensure_connected()
                    if self._session is None:
                        raise RuntimeError("official MCP session unavailable")
                    result = self._submit(self._session.call_tool(name, arguments=args))
                    normalized = self._normalize_result(result)
                    self._record_success()
                    return normalized
                except OfficialToolError as exc:
                    if exc.code == "official_tool_error":
                        # Do not degrade semantic tool errors into transport failures.
                        raise
                    self._record_failure(exc)
                    try:
                        self._submit(self._disconnect_async())
                    except Exception as cleanup_exc:
                        self._log_cleanup_exception("Official MCP disconnect failed", cleanup_exc)
                    if attempt == 1:
                        raise
                except Exception as exc:
                    self._record_failure(exc)
                    try:
                        self._submit(self._disconnect_async())
                    except Exception as cleanup_exc:
                        self._log_cleanup_exception("Official MCP disconnect failed", cleanup_exc)
                    if attempt == 1:
                        raise OfficialToolError(
                            "official_unavailable",
                            f"official MCP call failed for tool '{name}': {exc}",
                        ) from exc

        raise OfficialToolError("official_unavailable", "official MCP unavailable")

    def list_tools(self) -> list[str]:
        with self._lock:
            for attempt in range(2):
                try:
                    self._ensure_connected()
                    if self._session is None:
                        return []
                    result = self._submit(self._session.list_tools())
                    tools = getattr(result, "tools", []) or []
                    self._record_success()
                    return [t.name for t in tools if getattr(t, "name", None)]
                except Exception as exc:
                    self._record_failure(exc)
                    try:
                        self._submit(self._disconnect_async())
                    except Exception as cleanup_exc:
                        self._log_cleanup_exception("Official MCP disconnect failed", cleanup_exc)
                    if attempt == 1:
                        raise OfficialToolError(
                            "official_unavailable",
                            f"official MCP list_tools failed: {exc}",
                        ) from exc
            return []

    def get_health(self) -> dict[str, Any]:
        with self._lock:
            health: dict[str, Any] = {
                "transport": self._transport,
                "url": self._url,
                "connected": self._session is not None,
                "failureCount": self._failure_count,
                "lastError": self._last_error,
                "lastFailureAt": self._last_failure_at,
                "lastConnectedAt": self._last_connected_at,
            }
            if self._transport == "stdio":
                health["command"] = self._command
                health["args"] = self._args
            else:
                health["hasHeaders"] = self._headers is not None
            return health

    def close(self) -> None:
        with self._lock:
            loop = self._loop
            thread = self._thread

            if loop:
                try:
                    self._submit(self._disconnect_async())
                except Exception as exc:
                    self._log_cleanup_exception("Official MCP disconnect during close failed", exc)

                loop.call_soon_threadsafe(loop.stop)

            if thread and thread.is_alive():
                thread.join(timeout=1.0)

            if loop and (thread is None or not thread.is_alive()):
                loop.close()
            elif thread and thread.is_alive():
                logger.warning("Official MCP loop thread did not stop within timeout")

            self._session = None
            self._session_cm = None
            self._transport_cm = None
            self._loop = None
            self._thread = None



"""
Auto-detect Linear IndexedDB object store hashes by sampling records.

Linear uses hash-based object store names that may change between versions.
This module detects stores by examining the structure of their records.
"""

from dataclasses import dataclass
from typing import Any

from ccl_chromium_reader import ccl_chromium_indexeddb  # type: ignore


@dataclass
class DetectedStores:
    """Container for detected object store names."""

    issues: str | None = None
    teams: str | None = None
    users: list[str] | None = None
    workflow_states: list[str] | None = None
    comments: str | None = None
    projects: str | None = None
    issue_content: str | None = None  # Y.js encoded issue descriptions
    labels: list[str] | None = None  # Issue labels (team + workspace)
    initiatives: str | None = None
    project_statuses: str | None = None
    cycles: str | None = None
    documents: str | None = None
    document_content: str | None = None
    milestones: str | None = None
    project_updates: str | None = None


def _is_issue_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like an issue."""
    required = {"number", "teamId", "stateId", "title"}
    return required.issubset(record.keys())


def _is_user_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a user."""
    required = {"name", "displayName", "email"}
    return required.issubset(record.keys())


def _is_team_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a team."""
    if not {"key", "name"}.issubset(record.keys()):
        return False
    key = record.get("key")
    if not isinstance(key, str):
        return False
    return key.isupper() and key.isalpha() and len(key) <= 10


def _is_workflow_state_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a workflow state."""
    if not {"name", "type", "color", "teamId"}.issubset(record.keys()):
        return False
    state_type = record.get("type")
    valid_types = {"started", "unstarted", "completed", "canceled", "backlog"}
    return state_type in valid_types


def _is_comment_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a comment."""
    required = {"issueId", "userId", "bodyData", "createdAt"}
    return required.issubset(record.keys())


def _is_project_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a project."""
    required = {"name", "teamIds", "slugId", "statusId", "memberIds"}
    return required.issubset(record.keys())


def _is_issue_content_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like issue content (Y.js encoded description)."""
    required = {"issueId", "contentState"}
    return required.issubset(record.keys())


def _is_label_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a label."""
    required = {"name", "color", "isGroup"}
    return required.issubset(record.keys())


def _is_initiative_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like an initiative."""
    required = {"name", "ownerId", "slugId", "frequencyResolution"}
    return required.issubset(record.keys())


def _is_project_status_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a project status."""
    if not {"name", "color", "position", "type", "indefinite"}.issubset(record.keys()):
        return False
    # Must not have teamId (that's workflow state)
    return "teamId" not in record


def _is_cycle_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a cycle."""
    required = {"number", "teamId", "startsAt", "endsAt"}
    return required.issubset(record.keys())


def _is_document_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a document."""
    required = {"title", "slugId", "projectId", "sortOrder"}
    has_required = required.issubset(record.keys())
    # Must not be an issue
    not_issue = "number" not in record and "stateId" not in record
    return has_required and not_issue


def _is_document_content_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like document content."""
    required = {"documentContentId", "contentData"}
    return required.issubset(record.keys())


def _is_milestone_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a project milestone."""
    required = {"name", "projectId", "sortOrder"}
    has_required = required.issubset(record.keys())
    # May have targetDate, currentProgress
    has_progress = "currentProgress" in record or "targetDate" in record
    return has_required and has_progress


def _is_project_update_record(record: dict[str, Any]) -> bool:
    """Check if a record looks like a project update."""
    # Has body and either projectId or health field
    has_body = "body" in record
    has_project = "projectId" in record or "health" in record
    # Must not be a comment
    not_comment = "issueId" not in record
    return has_body and has_project and not_comment


def detect_stores(db: ccl_chromium_indexeddb.WrappedDatabase) -> DetectedStores:
    """
    Detect object stores by sampling their first record.

    Args:
        db: The wrapped IndexedDB database to scan.

    Returns:
        DetectedStores with detected store names for each entity type.
    """
    result = DetectedStores(users=[], workflow_states=[], labels=[])

    for store_name in db.object_store_names:
        if store_name is None or store_name.startswith("_") or "_partial" in store_name:
            continue

        try:
            store = db[store_name]
            for record in store.iterate_records():
                val = record.value
                if not isinstance(val, dict):
                    break

                if _is_issue_record(val) and result.issues is None:
                    result.issues = store_name
                elif _is_team_record(val) and result.teams is None:
                    result.teams = store_name
                elif _is_user_record(val) and store_name not in (result.users or []):
                    if result.users is None:
                        result.users = []
                    result.users.append(store_name)
                elif _is_workflow_state_record(val) and store_name not in (
                    result.workflow_states or []
                ):
                    if result.workflow_states is None:
                        result.workflow_states = []
                    result.workflow_states.append(store_name)
                elif _is_comment_record(val) and result.comments is None:
                    result.comments = store_name
                elif _is_project_record(val) and result.projects is None:
                    result.projects = store_name
                elif _is_issue_content_record(val) and result.issue_content is None:
                    result.issue_content = store_name
                elif _is_label_record(val) and store_name not in (result.labels or []):
                    if result.labels is None:
                        result.labels = []
                    result.labels.append(store_name)
                elif _is_initiative_record(val) and result.initiatives is None:
                    result.initiatives = store_name
                elif _is_project_status_record(val) and result.project_statuses is None:
                    result.project_statuses = store_name
                elif _is_cycle_record(val) and result.cycles is None:
                    result.cycles = store_name
                elif _is_document_record(val) and result.documents is None:
                    result.documents = store_name
                elif _is_document_content_record(val) and result.document_content is None:
                    result.document_content = store_name
                elif _is_milestone_record(val) and result.milestones is None:
                    result.milestones = store_name
                elif _is_project_update_record(val) and result.project_updates is None:
                    result.project_updates = store_name

                break  # Only check first record
        except Exception:
            continue

    return result



"""
Local tool handlers backed by Linear's IndexedDB cache.

These handlers contain only local-read logic. Routing/fallback decisions live in
`router.py`.
"""

from __future__ import annotations

import heapq
from typing import Any, Callable

from .reader import LinearLocalReader


class LocalFallbackRequested(RuntimeError):
    """Raised when local execution should fall back to the official MCP."""

    def __init__(self, code: str, message: str):
        super().__init__(message)
        self.code = code
        self.message = message


def _empty_total(key: str) -> dict[str, Any]:
    return {key: [], "totalCount": 0}


def _serialize_progress(progress: dict[str, Any] | None) -> dict[str, int] | None:
    if not progress:
        return None
    return {
        "completed": progress.get("completedIssueCount", 0),
        "started": progress.get("startedIssueCount", 0),
        "unstarted": progress.get("unstartedIssueCount", 0),
        "total": progress.get("scopeCount", 0),
    }


def _serialize_status_update(
    reader: LinearLocalReader, update: dict[str, Any]
) -> dict[str, Any]:
    return {
        "id": update.get("id"),
        "body": update.get("body"),
        "health": update.get("health"),
        "author": reader.get_user_name(update.get("userId")),
        "project": reader.get_project_name(update.get("projectId")),
        "createdAt": update.get("createdAt"),
        "updatedAt": update.get("updatedAt"),
    }


def _collect_status_updates(
    reader: LinearLocalReader,
    project_id: str | None = None,
    user_id: str | None = None,
    order_by: str = "createdAt",
) -> list[dict[str, Any]]:
    updates = list(reader.project_updates.values())

    if project_id:
        updates = [u for u in updates if u.get("projectId") == project_id]
    if user_id:
        updates = [u for u in updates if u.get("userId") == user_id]

    sort_key = "updatedAt" if order_by == "updatedAt" else "createdAt"
    updates.sort(key=lambda x: x.get(sort_key) or "", reverse=True)
    return updates


def list_issues(
    reader: LinearLocalReader,
    assignee: str | None = None,
    team: str | None = None,
    state: str | None = None,
    priority: int | None = None,
    project: str | None = None,
    query: str | None = None,
    orderBy: str = "updatedAt",
    limit: int = 50,
) -> dict[str, Any]:
    assignee_id = None
    if assignee:
        user = reader.find_user(assignee)
        if user:
            assignee_id = user["id"]
        else:
            return _empty_total("issues")

    team_id = None
    if team:
        team_obj = reader.find_team(team)
        if team_obj:
            team_id = team_obj["id"]
        else:
            return _empty_total("issues")

    project_id = None
    if project:
        project_obj = reader.find_project(project)
        if project_obj:
            project_id = project_obj["id"]
        else:
            return _empty_total("issues")

    query_lower = query.lower() if query else None
    state_lower = state.lower() if state else None
    sort_key = "createdAt" if orderBy == "createdAt" else "updatedAt"

    filtered: list[dict[str, Any]] = []
    for issue in reader.issues.values():
        if assignee_id and issue.get("assigneeId") != assignee_id:
            continue
        if team_id and issue.get("teamId") != team_id:
            continue
        if state_lower:
            issue_state_type = reader.get_state_type(issue.get("stateId", ""))
            issue_state_name = reader.get_state_name(issue.get("stateId", ""))
            if state_lower != issue_state_type and state_lower != (
                issue_state_name or ""
            ).lower():
                continue
        if project_id and issue.get("projectId") != project_id:
            continue
        if query_lower and query_lower not in (issue.get("title") or "").lower():
            continue
        if priority is not None and issue.get("priority") != priority:
            continue
        filtered.append(issue)

    total_count = len(filtered)
    if limit and limit > 0:
        page = heapq.nlargest(limit, filtered, key=lambda x: x.get(sort_key) or "")
    else:
        page = sorted(filtered, key=lambda x: x.get(sort_key) or "", reverse=True)

    results = []
    for issue in page:
        results.append(
            {
                "identifier": issue.get("identifier"),
                "title": issue.get("title"),
                "priority": issue.get("priority"),
                "state": reader.get_state_name(issue.get("stateId", "")),
                "stateType": reader.get_state_type(issue.get("stateId", "")),
                "assignee": reader.get_user_name(issue.get("assigneeId")),
                "dueDate": issue.get("dueDate"),
            }
        )

    return {"issues": results, "totalCount": total_count}


def get_issue(reader: LinearLocalReader, id: str) -> dict[str, Any] | None:
    issue = reader.get_issue_by_identifier(id)
    if not issue:
        return None

    comments = reader.get_comments_for_issue(issue["id"])
    enriched_comments = []
    for comment in comments:
        user = reader.users.get(comment.get("userId", ""), {})
        enriched_comments.append(
            {
                "author": user.get("name", "Unknown"),
                "body": comment.get("body", ""),
                "createdAt": comment.get("createdAt"),
            }
        )

    return {
        "identifier": issue.get("identifier"),
        "title": issue.get("title"),
        "description": issue.get("description"),
        "priority": issue.get("priority"),
        "estimate": issue.get("estimate"),
        "state": reader.get_state_name(issue.get("stateId", "")),
        "stateType": reader.get_state_type(issue.get("stateId", "")),
        "assignee": reader.get_user_name(issue.get("assigneeId")),
        "project": reader.get_project_name(issue.get("projectId")),
        "dueDate": issue.get("dueDate"),
        "createdAt": issue.get("createdAt"),
        "updatedAt": issue.get("updatedAt"),
        "comments": enriched_comments,
        "url": f"https://linear.app/issue/{issue.get('identifier')}",
    }


def list_teams(reader: LinearLocalReader) -> list[dict[str, Any]]:
    results = []
    for team in reader.teams.values():
        results.append(
            {
                "key": team.get("key"),
                "name": team.get("name"),
                "issueCount": reader.get_issue_count_for_team(team.get("id")),
            }
        )
    results.sort(key=lambda x: x.get("key", ""))
    return results


def list_projects(
    reader: LinearLocalReader, team: str | None = None
) -> list[dict[str, Any]]:
    team_id = None
    if team:
        team_obj = reader.find_team(team)
        if team_obj:
            team_id = team_obj["id"]
        else:
            return []

    results = []
    for project in reader.projects.values():
        if team_id and team_id not in project.get("teamIds", []):
            continue

        results.append(
            {
                "name": project.get("name"),
                "state": project.get("state"),
                "issueCount": reader.get_issue_count_for_project(project.get("id")),
                "startDate": project.get("startDate"),
                "targetDate": project.get("targetDate"),
            }
        )

    results.sort(key=lambda x: x.get("name", "") or "")
    return results


def get_team(reader: LinearLocalReader, query: str) -> dict[str, Any] | None:
    team_obj = reader.find_team(query)
    if not team_obj:
        return None

    return {
        "id": team_obj.get("id"),
        "key": team_obj.get("key"),
        "name": team_obj.get("name"),
        "description": team_obj.get("description"),
        "issueCount": reader.get_issue_count_for_team(team_obj.get("id")),
        "issuesByState": reader.get_issue_state_counts_for_team(team_obj.get("id")),
    }


def get_project(reader: LinearLocalReader, query: str) -> dict[str, Any] | None:
    project = reader.find_project(query)
    if not project:
        return None

    return {
        "id": project.get("id"),
        "name": project.get("name"),
        "description": project.get("description"),
        "state": project.get("state"),
        "startDate": project.get("startDate"),
        "targetDate": project.get("targetDate"),
        "issueCount": reader.get_issue_count_for_project(project.get("id")),
        "issuesByState": reader.get_issue_state_counts_for_project(project.get("id")),
    }


def list_users(reader: LinearLocalReader) -> list[dict[str, Any]]:
    results = []
    for user in reader.users.values():
        results.append(
            {
                "id": user.get("id"),
                "name": user.get("name"),
                "email": user.get("email"),
                "displayName": user.get("displayName"),
                "assignedIssueCount": reader.get_issue_count_for_user(user.get("id")),
            }
        )
    results.sort(key=lambda x: x.get("name", "") or "")
    return results


def get_user(reader: LinearLocalReader, query: str) -> dict[str, Any] | None:
    user = reader.find_user(query)
    if not user:
        return None

    state_counts = reader.get_issue_state_counts_for_user(user.get("id"))
    return {
        "id": user.get("id"),
        "name": user.get("name"),
        "email": user.get("email"),
        "displayName": user.get("displayName"),
        "assignedIssueCount": sum(state_counts.values()),
        "issuesByState": state_counts,
    }


def list_issue_statuses(reader: LinearLocalReader, team: str) -> list[dict[str, Any]]:
    team_obj = reader.find_team(team)
    if not team_obj:
        return []

    results = []
    for state in reader.states.values():
        if state.get("teamId") == team_obj["id"]:
            results.append(
                {
                    "id": state.get("id"),
                    "name": state.get("name"),
                    "type": state.get("type"),
                    "color": state.get("color"),
                    "position": state.get("position"),
                }
            )

    results.sort(key=lambda x: (x.get("position") or 0))
    return results


def get_issue_status(
    reader: LinearLocalReader,
    team: str,
    name: str | None = None,
    id: str | None = None,
) -> dict[str, Any] | None:
    team_obj = reader.find_team(team)
    if not team_obj:
        return None

    query = id or name
    if not query:
        return None

    status = reader.find_issue_status(team_obj["id"], query)
    if not status:
        return None

    return {
        "id": status.get("id"),
        "name": status.get("name"),
        "type": status.get("type"),
        "color": status.get("color"),
        "position": status.get("position"),
        "team": team_obj.get("name"),
    }


def list_comments(reader: LinearLocalReader, issueId: str) -> list[dict[str, Any]]:
    issue = reader.get_issue_by_identifier(issueId)
    if not issue:
        return []

    comments = reader.get_comments_for_issue(issue["id"])
    results = []
    for comment in comments:
        user = reader.users.get(comment.get("userId", ""), {})
        results.append(
            {
                "id": comment.get("id"),
                "author": user.get("name", "Unknown"),
                "body": comment.get("body", ""),
                "createdAt": comment.get("createdAt"),
                "updatedAt": comment.get("updatedAt"),
            }
        )

    return results


def list_issue_labels(
    reader: LinearLocalReader, team: str | None = None
) -> list[dict[str, Any]]:
    team_id = None
    if team:
        team_obj = reader.find_team(team)
        if team_obj:
            team_id = team_obj["id"]

    results = []
    for label in reader.labels.values():
        if team_id and label.get("teamId") and label.get("teamId") != team_id:
            continue
        results.append(
            {
                "id": label.get("id"),
                "name": label.get("name"),
                "color": label.get("color"),
                "isGroup": label.get("isGroup"),
            }
        )

    results.sort(key=lambda x: x.get("name", "") or "")
    return results


def list_initiatives(reader: LinearLocalReader) -> list[dict[str, Any]]:
    results = []
    for initiative in reader.initiatives.values():
        results.append(
            {
                "id": initiative.get("id"),
                "name": initiative.get("name"),
                "slugId": initiative.get("slugId"),
                "color": initiative.get("color"),
                "status": initiative.get("status"),
                "owner": reader.get_user_name(initiative.get("ownerId")),
            }
        )

    results.sort(key=lambda x: x.get("name", "") or "")
    return results


def get_initiative(reader: LinearLocalReader, query: str) -> dict[str, Any] | None:
    initiative = reader.find_initiative(query)
    if not initiative:
        return None

    return {
        "id": initiative.get("id"),
        "name": initiative.get("name"),
        "slugId": initiative.get("slugId"),
        "color": initiative.get("color"),
        "status": initiative.get("status"),
        "owner": reader.get_user_name(initiative.get("ownerId")),
        "teamIds": initiative.get("teamIds", []),
        "createdAt": initiative.get("createdAt"),
        "updatedAt": initiative.get("updatedAt"),
    }


def list_cycles(reader: LinearLocalReader, teamId: str) -> list[dict[str, Any]]:
    team_obj = reader.find_team(teamId)
    if not team_obj:
        return []

    cycles = reader.get_cycles_for_team(team_obj["id"])
    results = []
    for cycle in cycles:
        results.append(
            {
                "id": cycle.get("id"),
                "number": cycle.get("number"),
                "startsAt": cycle.get("startsAt"),
                "endsAt": cycle.get("endsAt"),
                "completedAt": cycle.get("completedAt"),
                "progress": _serialize_progress(cycle.get("currentProgress")),
            }
        )

    return results


def list_documents(
    reader: LinearLocalReader, project: str | None = None
) -> list[dict[str, Any]]:
    project_id = None
    if project:
        project_obj = reader.find_project(project)
        if project_obj:
            project_id = project_obj["id"]
        else:
            return []

    results = []
    for doc in reader.documents.values():
        if project_id and doc.get("projectId") != project_id:
            continue
        results.append(
            {
                "id": doc.get("id"),
                "title": doc.get("title"),
                "slugId": doc.get("slugId"),
                "project": reader.get_project_name(doc.get("projectId")),
                "createdAt": doc.get("createdAt"),
                "updatedAt": doc.get("updatedAt"),
            }
        )

    results.sort(key=lambda x: x.get("updatedAt", "") or "", reverse=True)
    return results


def get_document(reader: LinearLocalReader, id: str) -> dict[str, Any] | None:
    doc = reader.find_document(id)
    if not doc:
        return None

    return {
        "id": doc.get("id"),
        "title": doc.get("title"),
        "slugId": doc.get("slugId"),
        "project": reader.get_project_name(doc.get("projectId")),
        "creator": reader.get_user_name(doc.get("creatorId")),
        "createdAt": doc.get("createdAt"),
        "updatedAt": doc.get("updatedAt"),
        "url": f"https://linear.app/document/{doc.get('slugId')}",
    }


def list_milestones(reader: LinearLocalReader, project: str) -> list[dict[str, Any]]:
    project_obj = reader.find_project(project)
    if not project_obj:
        return []

    milestones = reader.get_milestones_for_project(project_obj["id"])
    results = []
    for milestone in milestones:
        results.append(
            {
                "id": milestone.get("id"),
                "name": milestone.get("name"),
                "targetDate": milestone.get("targetDate"),
                "progress": _serialize_progress(milestone.get("currentProgress")),
            }
        )

    return results


def get_milestone(
    reader: LinearLocalReader, project: str, query: str
) -> dict[str, Any] | None:
    project_obj = reader.find_project(project)
    if not project_obj:
        return None

    milestone = reader.find_milestone(project_obj["id"], query)
    if not milestone:
        return None

    return {
        "id": milestone.get("id"),
        "name": milestone.get("name"),
        "project": project_obj.get("name"),
        "targetDate": milestone.get("targetDate"),
        "sortOrder": milestone.get("sortOrder"),
        "progress": _serialize_progress(milestone.get("currentProgress")),
    }


def get_status_updates(
    reader: LinearLocalReader,
    type: str,
    id: str | None = None,
    project: str | None = None,
    initiative: str | None = None,
    user: str | None = None,
    includeArchived: bool | None = None,
    orderBy: str = "createdAt",
    limit: int = 50,
    cursor: str | None = None,
    createdAt: str | None = None,
    updatedAt: str | None = None,
) -> dict[str, Any] | None:
    if type != "project":
        raise LocalFallbackRequested(
            "unsupported_type",
            "local cache supports only get_status_updates(type='project')",
        )

    if initiative or cursor or createdAt or updatedAt or includeArchived:
        raise LocalFallbackRequested(
            "unsupported_filter",
            "one or more filters are unsupported by local cache",
        )

    project_id = None
    if project:
        project_obj = reader.find_project(project)
        if not project_obj:
            return _empty_total("statusUpdates")
        project_id = project_obj["id"]

    user_id = None
    if user:
        user_obj = reader.find_user(user)
        if not user_obj:
            return _empty_total("statusUpdates")
        user_id = user_obj["id"]

    updates = _collect_status_updates(reader, project_id, user_id, orderBy)
    if id:
        for update in updates:
            if update.get("id") == id:
                return _serialize_status_update(reader, update)
        return None

    total_count = len(updates)
    if limit and limit > 0:
        updates = updates[:limit]

    return {
        "statusUpdates": [_serialize_status_update(reader, u) for u in updates],
        "totalCount": total_count,
    }


def list_project_updates(reader: LinearLocalReader, project: str) -> list[dict[str, Any]]:
    project_obj = reader.find_project(project)
    if not project_obj:
        return []

    response = get_status_updates(reader, type="project", project=project, limit=0)
    if not isinstance(response, dict):
        return []
    return response.get("statusUpdates", [])


LOCAL_READ_HANDLERS: dict[str, Callable[..., Any]] = {
    "list_issues": list_issues,
    "get_issue": get_issue,
    "list_teams": list_teams,
    "list_projects": list_projects,
    "get_team": get_team,
    "get_project": get_project,
    "list_users": list_users,
    "get_user": get_user,
    "list_issue_statuses": list_issue_statuses,
    "get_issue_status": get_issue_status,
    "list_comments": list_comments,
    "list_issue_labels": list_issue_labels,
    "list_initiatives": list_initiatives,
    "get_initiative": get_initiative,
    "list_cycles": list_cycles,
    "list_documents": list_documents,
    "get_document": get_document,
    "list_milestones": list_milestones,
    "get_milestone": get_milestone,
    "get_status_updates": get_status_updates,
    "list_project_updates": list_project_updates,
}




"""
Copyright 2022, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import io
import datetime
import struct
import os


__version__ = "0.1"
__description__ = "Module for reading Chromium Pickles."
__contact__ = "Alex Caithness"


class EasyPickleException(Exception):
    ...


class EasyPickleIterator:
    """
    A pythonic implementation of the PickleIterator object used in various places in Chrom(e|ium).
    """
    def __init__(self, data: bytes, alignment: int=4):
        """
        Takes a bytes buffer and wraps the EasyPickleIterator around it
        :param data: the data to be wrapped
        :param alignment: (optional) the number of bytes to align reads to (default: 4)
        """
        self._f = io.BytesIO(data)
        self._alignment = alignment

        self._pickle_length = self.read_uint32()
        if len(data) != self._pickle_length + 4:
            raise EasyPickleException("pickle length invalid")

    def __enter__(self) -> "EasyPickleIterator":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._f.close()

    def read_aligned(self, length: int) -> bytes:
        """
        reads the number of bytes specified by the length parameter. Aligns the buffer afterwards if required.
        :param length: the length od data to be read
        :return: the data read (without the alignment padding)
        """
        raw = self._f.read(length)
        if len(raw) != length:
            raise EasyPickleException(f"Tried to read {length} bytes but only got {len(raw)}")

        align_count = self._alignment - (length % self._alignment)
        if align_count != self._alignment:
            self._f.seek(align_count, os.SEEK_CUR)

        return raw

    def read_uint16(self) -> int:
        raw = self.read_aligned(2)
        return struct.unpack("<H", raw)[0]

    def read_uint32(self) -> int:
        raw = self.read_aligned(4)
        return struct.unpack("<I", raw)[0]

    def read_uint64(self) -> int:
        raw = self.read_aligned(8)
        return struct.unpack("<Q", raw)[0]

    def read_int16(self) -> int:
        raw = self.read_aligned(2)
        return struct.unpack("<h", raw)[0]

    def read_int32(self) -> int:
        raw = self.read_aligned(4)
        return struct.unpack("<i", raw)[0]

    def read_int64(self) -> int:
        raw = self.read_aligned(8)
        return struct.unpack("<q", raw)[0]

    def read_bool(self) -> bool:
        raw = self.read_int32()
        if raw == 0:
            return False
        elif raw == 1:
            return True
        else:
            raise EasyPickleException("bools should only contain 0 or 1")

    def read_single(self) -> float:
        raw = self.read_aligned(4)
        return struct.unpack("<f", raw)[0]

    def read_double(self) -> float:
        raw = self.read_aligned(8)
        return struct.unpack("<d", raw)[0]

    def read_string(self) -> str:
        length = self.read_uint32()
        raw = self.read_aligned(length)
        return raw.decode("utf-8")

    def read_string16(self) -> str:
        length = self.read_uint32() * 2  # character count
        raw = self.read_aligned(length)
        return raw.decode("utf-16-le")

    def read_datetime(self) -> datetime.datetime:
        return datetime.datetime(1601, 1, 1) + datetime.timedelta(microseconds=self.read_uint64())




"""
Copyright 2020, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import enum
import typing
from dataclasses import dataclass

from . import ccl_v8_value_deserializer

# See: https://chromium.googlesource.com/chromium/src/third_party/+/master/blink/renderer/bindings/core/v8/serialization
#      https://source.chromium.org/chromium/chromium/src/+/main:third_party/blink/renderer/bindings/modules/v8/serialization/v8_script_value_serializer_for_modules.cc


# WebCoreStrings are read as (length:uint32_t, string:UTF8[length]).
# RawStrings are read as (length:uint32_t, string:UTF8[length]).
# RawUCharStrings are read as
#     (length:uint32_t, string:UChar[length/sizeof(UChar)]).
# RawFiles are read as
#     (path:WebCoreString, url:WebCoreStrng, type:WebCoreString).
# There is a reference table that maps object references (uint32_t) to
# v8::Values.
# Tokens marked with (ref) are inserted into the reference table and given the
# next object reference ID after decoding.
# All tags except InvalidTag, PaddingTag, ReferenceCountTag, VersionTag,
# GenerateFreshObjectTag and GenerateFreshArrayTag push their results to the
# deserialization stack.
# There is also an 'open' stack that is used to resolve circular references.
# Objects or arrays may contain self-references. Before we begin to deserialize
# the contents of these values, they are first given object reference IDs (by
# GenerateFreshObjectTag/GenerateFreshArrayTag); these reference IDs are then
# used with ObjectReferenceTag to tie the recursive knot.

__version__ = "0.3"
__description__ = "Partial reimplementation of the Blink Javascript Object Serialization"
__contact__ = "Alex Caithness"

__DEBUG = False


def log(msg, debug_only=True):
    if __DEBUG or not debug_only:
        caller_name = sys._getframe(1).f_code.co_name
        caller_line = sys._getframe(1).f_code.co_firstlineno
        print(f"{caller_name} ({caller_line}):\t{msg}")


class BlobIndexType(enum.Enum):
    Blob = 0
    File = 1


@dataclass
class BlobIndex:
    index_type: BlobIndexType
    index_id: int


@dataclass(frozen=True)
class NativeFileHandle:
    is_dir: bool
    name: str
    token_index: int


@dataclass(frozen=True)
class CryptoKey:
    sub_type: "V8CryptoKeySubType"
    algorithm_type: typing.Optional["V8CryptoKeyAlgorithm"]
    hash_type: typing.Optional["V8CryptoKeyAlgorithm"]
    asymmetric_key_type: typing.Optional["V8AsymmetricCryptoKeyType"]
    byte_length: typing.Optional[int]
    public_exponent: typing.Optional[bytes]
    named_curve_type: typing.Optional["V8CryptoNamedCurve"]
    key_usage: "V8CryptoKeyUsage"
    key_data: bytes


class Constants:
    tag_kMessagePortTag = b"M"  # index:int -> MessagePort. Fills the result with
                                # transferred MessagePort.
    tag_kMojoHandleTag = b"h"   # index:int -> MojoHandle. Fills the result with
                                # transferred MojoHandle.
    tag_kBlobTag = b"b"         # uuid:WebCoreString, type:WebCoreString, size:uint64_t ->
                                # Blob (ref)
    tag_kBlobIndexTag = b"i"    # index:int32_t -> Blob (ref)
    tag_kFileTag = b"f"         # file:RawFile -> File (ref)
    tag_kFileIndexTag = b"e"    # index:int32_t -> File (ref)
    tag_kDOMFileSystemTag = b"d"  # type : int32_t, name:WebCoreString,
                                  # uuid:WebCoreString -> FileSystem (ref)
    tag_kNativeFileSystemFileHandleTag = b"n"  # name:WebCoreString, index:uint32_t
                                               # -> NativeFileSystemFileHandle (ref)
    tag_kNativeFileSystemDirectoryHandleTag = b"N"  # name:WebCoreString, index:uint32_t ->
                                                   # NativeFileSystemDirectoryHandle (ref)
    tag_kFileListTag = b"l"                     # length:uint32_t, files:RawFile[length] -> FileList (ref)
    tag_kFileListIndexTag = b"L"                # length:uint32_t, files:int32_t[length] -> FileList (ref)
    tag_kImageDataTag = b"#"                   # tags terminated by ImageSerializationTag::kEnd (see
                                               # SerializedColorParams.h), width:uint32_t,
                                               # height:uint32_t, pixelDataLength:uint64_t,
                                               # data:byte[pixelDataLength]
                                               # -> ImageData (ref)
    tag_kImageBitmapTag = b"g"        # tags terminated by ImageSerializationTag::kEnd (see
                                      # SerializedColorParams.h), width:uint32_t,
                                      # height:uint32_t, pixelDataLength:uint32_t,
                                      # data:byte[pixelDataLength]
                                      # -> ImageBitmap (ref)
    tag_kImageBitmapTransferTag = b"G"      # index:uint32_t -> ImageBitmap. For ImageBitmap transfer
    tag_kOffscreenCanvasTransferTag = b"H"  # index, width, height, id,
                                            # filter_quality::uint32_t ->
                                            # OffscreenCanvas. For OffscreenCanvas
                                            # transfer
    tag_kReadableStreamTransferTag = b"r"    # index:uint32_t
    tag_kTransformStreamTransferTag = b"m"   # index:uint32_t
    tag_kWritableStreamTransferTag = b"w"    # index:uint32_t
    tag_kDOMPointTag = b"Q"                  # x:Double, y:Double, z:Double, w:Double
    tag_kDOMPointReadOnlyTag = b"W"          # x:Double, y:Double, z:Double, w:Double
    tag_kDOMRectTag = b"E"                   # x:Double, y:Double, width:Double, height:Double
    tag_kDOMRectReadOnlyTag = b"R"           # x:Double, y:Double, width:Double, height:Double
    tag_kDOMQuadTag = b"T"                   # p1:Double, p2:Double, p3:Double, p4:Double
    tag_kDOMMatrixTag = b"Y"                 # m11..m44: 16 Double
    tag_kDOMMatrixReadOnlyTag = b"U"         # m11..m44: 16 Double
    tag_kDOMMatrix2DTag = b"I"               # a..f: 6 Double
    tag_kDOMMatrix2DReadOnlyTag = b"O"       # a..f: 6 Double
    tag_kCryptoKeyTag = b"K"                 # subtag:byte, props, usages:uint32_t,
    # keyDataLength:uint32_t, keyData:byte[keyDataLength]
    #                 If subtag=AesKeyTag:
    #                     props = keyLengthBytes:uint32_t, algorithmId:uint32_t
    #                 If subtag=HmacKeyTag:
    #                     props = keyLengthBytes:uint32_t, hashId:uint32_t
    #                 If subtag=RsaHashedKeyTag:
    #                     props = algorithmId:uint32_t, type:uint32_t,
    #                     modulusLengthBits:uint32_t,
    #                     publicExponentLength:uint32_t,
    #                     publicExponent:byte[publicExponentLength],
    #                     hashId:uint32_t
    #                 If subtag=EcKeyTag:
    #                     props = algorithmId:uint32_t, type:uint32_t,
    #                     namedCurve:uint32_t
    tag_kRTCCertificateTag = b"k"  # length:uint32_t, pemPrivateKey:WebCoreString,
    # pemCertificate:WebCoreString
    tag_kRTCEncodedAudioFrameTag = b"A"  # uint32_t -> transferred audio frame ID
    tag_kRTCEncodedVideoFrameTag = b"V"  # uint32_t -> transferred video frame ID
    tag_kVideoFrameTag = b"v"            # uint32_t -> transferred video frame ID

    # The following tags were used by the Shape Detection API implementation
    # between M71 and M81. During these milestones, the API was always behind
    # a flag. Usage was removed in https:#crrev.com/c/2040378.
    tag_kDeprecatedDetectedBarcodeTag = b"B"
    tag_kDeprecatedDetectedFaceTag = b"F"
    tag_kDeprecatedDetectedTextTag = b"t"

    tag_kDOMExceptionTag = b"x"  # name:String,message:String,stack:String
    tag_kVersionTag = b"\xff"  # version:uint32_t -> Uses this as the file version.
    tag_kTrailerOffsetTag = b"\xfe" # offset:uint64_t (fixed width, network order) from buffer, start size:uint32_t (fixed width, network order)
    tag_kTrailerRequiresInterfacesTag = b"\xA0"


class V8CryptoKeySubType(enum.IntEnum):
    """
    See: third_party/blink/renderer/bindings/modules/v8/serialization/web_crypto_sub_tags.h
    Used by the kCryptoKeyTag type
    """
    AesKey = 1
    HmacKey = 2
    # ID 3 was used by RsaKeyTag, while still behind experimental flag.
    RsaHashedKey = 4
    EcKey = 5
    NoParamsKey = 6


class V8CryptoKeyAlgorithm(enum.IntEnum):
    """
    See: third_party/blink/renderer/bindings/modules/v8/serialization/web_crypto_sub_tags.h
    Used by the kCryptoKeyTag type
    """
    AesCbcTag = 1
    HmacTag = 2
    RsaSsaPkcs1v1_5Tag = 3
    # ID 4 was used by RsaEs, while still behind experimental flag.
    Sha1Tag = 5
    Sha256Tag = 6
    Sha384Tag = 7
    Sha512Tag = 8
    AesGcmTag = 9
    RsaOaepTag = 10
    AesCtrTag = 11
    AesKwTag = 12
    RsaPssTag = 13
    EcdsaTag = 14
    EcdhTag = 15
    HkdfTag = 16
    Pbkdf2Tag = 17


class V8AsymmetricCryptoKeyType(enum.IntEnum):
    Public = 1
    Private = 2


class V8CryptoNamedCurve(enum.IntEnum):
    """
    See: third_party/blink/renderer/bindings/modules/v8/serialization/web_crypto_sub_tags.h
    Used by the kCryptoKeyTag type
    """
    P256 = 1
    P384 = 2
    P521 = 3


class V8CryptoKeyUsage(enum.IntFlag):
    """
    See: third_party/blink/renderer/bindings/modules/v8/serialization/web_crypto_sub_tags.h
    Used by the kCryptoKeyTag type
    """
    kExtractableUsage = 1 << 0
    kEncryptUsage = 1 << 1
    kDecryptUsage = 1 << 2
    kSignUsage = 1 << 3
    kVerifyUsage = 1 << 4
    kDeriveKeyUsage = 1 << 5
    kWrapKeyUsage = 1 << 6
    kUnwrapKeyUsage = 1 << 7
    kDeriveBitsUsage = 1 << 8


class BlinkV8Deserializer:
    def _read_varint(self, stream) -> int:
        return ccl_v8_value_deserializer.read_le_varint(stream)[0]

    def _read_varint32(self, stream) -> int:
        return ccl_v8_value_deserializer.read_le_varint(stream, is_32bit=True)[0]

    def _read_utf8_string(self, stream: typing.BinaryIO) -> str:
        length = self._read_varint32(stream)
        raw_string = stream.read(length)
        if len(raw_string) != length:
            raise ValueError("Could not read all of the utf-8 data")
        return raw_string.decode("utf-8")

    # def _read_uint32(self, stream: typing.BinaryIO) -> int:
    #     raw = stream.read(4)
    #     if len(raw) < 4:
    #         raise ValueError("Could not read enough data when reading int32")
    #     return struct.unpack("<I", raw)[0]

    def _read_file_index(self, stream: typing.BinaryIO) -> BlobIndex:
        return BlobIndex(BlobIndexType.File, self._read_varint(stream))

    def _read_blob_index(self, stream: typing.BinaryIO) -> BlobIndex:
        return BlobIndex(BlobIndexType.Blob, self._read_varint(stream))

    def _read_file_list_index(self, stream: typing.BinaryIO) -> typing.Iterable[BlobIndex]:
        length = self._read_varint(stream)
        result = [self._read_file_index(stream) for _ in range(length)]
        return result

    def _read_native_file_handle(self, is_dir: bool, stream: typing.BinaryIO) -> NativeFileHandle:
        return NativeFileHandle(is_dir, self._read_utf8_string(stream), self._read_varint(stream))

    def _read_crypto_key(self, stream: typing.BinaryIO):
        sub_type = V8CryptoKeySubType(stream.read(1)[0])

        if sub_type == V8CryptoKeySubType.AesKey:
            algorithm_id = V8CryptoKeyAlgorithm(self._read_varint32(stream))
            byte_length = self._read_varint32(stream)
            params = {
                "algorithm_type": algorithm_id,
                "byte_length": byte_length,
                "hash_type": None,
                "named_curve_type": None,
                "asymmetric_key_type": None,
                "public_exponent": None
            }
        elif sub_type == V8CryptoKeySubType.HmacKey:
            byte_length = self._read_varint32(stream)
            hash_id = V8CryptoKeyAlgorithm(self._read_varint32(stream))
            params = {
                "byte_length": byte_length,
                "hash_type": hash_id,
                "algorithm_type": None,
                "named_curve_type": None,
                "asymmetric_key_type": None,
                "public_exponent": None
            }
        elif sub_type == V8CryptoKeySubType.RsaHashedKey:
            algorithm_id = V8CryptoKeyAlgorithm(self._read_varint32(stream))
            asymmetric_key_type = V8AsymmetricCryptoKeyType(stream.read(1)[0])
            length_bytes = self._read_varint32(stream)
            public_exponent_length = self._read_varint32(stream)
            public_exponent = stream.read(public_exponent_length)
            if len(public_exponent) != public_exponent_length:
                raise ValueError(f"Could not read all of public exponent data")
            hash_id = V8CryptoKeyAlgorithm(self._read_varint32(stream))
            params = {
                "algorithm_type": algorithm_id,
                "asymmetric_key_type": asymmetric_key_type,
                "byte_length": length_bytes,
                "public_exponent": public_exponent,
                "hash_type": hash_id,
                "named_curve_type": None
            }

        elif sub_type == V8CryptoKeySubType.EcKey:
            algorithm_id = V8CryptoKeyAlgorithm(self._read_varint32(stream))
            asymmetric_key_type = V8AsymmetricCryptoKeyType(stream.read(1)[0])
            named_curve = V8CryptoNamedCurve(self._read_varint32(stream))
            params = {
                "algorithm_type": algorithm_id,
                "asymmetric_key_type": asymmetric_key_type,
                "named_curve_type": named_curve,
                "hash_type": None,
                "byte_length": None,
                "public_exponent": None
            }
        elif sub_type == V8CryptoKeySubType.NoParamsKey:
            algorithm_id = V8CryptoKeyAlgorithm(self._read_varint32(stream))
            params = {
                "algorithm_type": algorithm_id,
                "hash_type": None,
                "asymmetric_key_type": None,
                "byte_length": None,
                "named_curve_type": None,
                "public_exponent": None
            }
        else:
            raise ValueError(f"Unknown V8CryptoKeySubType {sub_type}")

        params["key_usage"] = V8CryptoKeyUsage(self._read_varint32(stream))
        key_length = self._read_varint32(stream)
        key_data = stream.read(key_length)
        if len(key_data) < key_length:
            raise ValueError("Could not read all key data")

        params["key_data"] = key_data
        return CryptoKey(sub_type, **params)

    def _not_implemented(self, stream):
        raise NotImplementedError()

    def read(self, stream: typing.BinaryIO) -> typing.Any:
        tag = stream.read(1)

        func = {
            Constants.tag_kMessagePortTag: lambda x: self._not_implemented(x),
            Constants.tag_kMojoHandleTag: lambda x: self._not_implemented(x),
            Constants.tag_kBlobTag: lambda x: self._not_implemented(x),
            Constants.tag_kBlobIndexTag: lambda x: self._read_blob_index(x),
            Constants.tag_kFileTag: lambda x: self._not_implemented(x),
            Constants.tag_kFileIndexTag: lambda x: self._read_file_index(x),
            Constants.tag_kDOMFileSystemTag: lambda x: self._not_implemented(x),
            Constants.tag_kNativeFileSystemFileHandleTag: lambda x: self._read_native_file_handle(False, x),
            Constants.tag_kNativeFileSystemDirectoryHandleTag: lambda x: self._read_native_file_handle(True, x),
            Constants.tag_kFileListTag: lambda x: self._not_implemented(x),
            Constants.tag_kFileListIndexTag: lambda x: self._read_file_list_index(x),
            Constants.tag_kImageDataTag: lambda x: self._not_implemented(x),
            Constants.tag_kImageBitmapTag: lambda x: self._not_implemented(x),
            Constants.tag_kImageBitmapTransferTag: lambda x: self._not_implemented(x),
            Constants.tag_kOffscreenCanvasTransferTag: lambda x: self._not_implemented(x),
            Constants.tag_kReadableStreamTransferTag: lambda x: self._not_implemented(x),
            Constants.tag_kTransformStreamTransferTag: lambda x: self._not_implemented(x),
            Constants.tag_kWritableStreamTransferTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMPointTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMPointReadOnlyTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMRectTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMRectReadOnlyTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMQuadTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMMatrixTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMMatrixReadOnlyTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMMatrix2DTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMMatrix2DReadOnlyTag: lambda x: self._not_implemented(x),
            Constants.tag_kCryptoKeyTag: lambda x: self._read_crypto_key(x),
            Constants.tag_kRTCCertificateTag: lambda x: self._not_implemented(x),
            Constants.tag_kRTCEncodedAudioFrameTag: lambda x: self._not_implemented(x),
            Constants.tag_kRTCEncodedVideoFrameTag: lambda x: self._not_implemented(x),
            Constants.tag_kVideoFrameTag: lambda x: self._not_implemented(x),
            Constants.tag_kDOMExceptionTag: lambda x: self._not_implemented(x)
        }.get(tag)

        if func is None:
            raise ValueError(f"Unknown tag: {tag}")

        return func(stream)



import sys
import pathlib
from ccl_chromium_reader import ccl_chromium_indexeddb
import time


def main(args):
    start = time.time()
    ldb_path = pathlib.Path(args[0])
    wrapper = ccl_chromium_indexeddb.WrappedIndexDB(ldb_path)

    for db_info in wrapper.database_ids:
        db = wrapper[db_info.dbid_no]
        print("------Database------")
        print(f"db_number={db.db_number}; name={db.name}; origin={db.origin}")
        print()
        print("\t---Object Stores---")
        for obj_store_name in db.object_store_names:
            obj_store = db[obj_store_name]
            print(f"\tobject_store_id={obj_store.object_store_id}; name={obj_store.name}")
            try:
                one_record = next(obj_store.iterate_records())
            except StopIteration:
                one_record = None
        print()
    end = time.time()
    print("Elapsed time: {} seconds.".format(int(end-start)))


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(f"USAGE: {pathlib.Path(sys.argv[0]).name} <ldb dir path>")
        exit(1)

    main(sys.argv[1:])



import re
import typing
import collections.abc as col_abc


KeySearch = typing.Union[str, re.Pattern, col_abc.Collection[str], col_abc.Callable[[str], bool]]


def is_keysearch_hit(search: KeySearch, value: str):
    if isinstance(search, str):
        return value == search
    elif isinstance(search, re.Pattern):
        return search.search(value) is not None
    elif isinstance(search, col_abc.Collection):
        return value in set(search)
    elif isinstance(search, col_abc.Callable):
        return search(value)
    else:
        raise TypeError(f"Unexpected type: {type(search)} (expects: {KeySearch})")


"""
Copyright 2022, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import dataclasses
import enum
import struct
import sys
import os
import pathlib
import datetime
import types
import typing
from .serialization_formats.ccl_easy_chromium_pickle import EasyPickleIterator, EasyPickleException

__version__ = "0.2"
__description__ = "Module for reading Chromium SNSS files"
__contact__ = "Alex Caithness"


class TabRestoreIdType(enum.Enum):
    # components/sessions/core/tab_restore_service_impl.cc
    CommandUpdateTabNavigation = 1
    CommandRestoredEntry = 2
    CommandWindowDeprecated = 3
    CommandSelectedNavigationInTab = 4
    CommandPinnedState = 5
    CommandSetExtensionAppID = 6
    CommandSetWindowAppName = 7
    CommandSetTabUserAgentOverride = 8
    CommandWindow = 9
    CommandSetTabGroupData = 10
    CommandSetTabUserAgentOverride2 = 11
    CommandSetWindowUserTitle = 12
    CommandCreateGroup = 13
    CommandAddTabExtraData = 14

    UnusedCommand = 255


class SessionRestoreIdType(enum.Enum):
    # components/sessions/core/session_service_commands.cc
    CommandSetTabWindow = 0
    CommandSetWindowBounds = 1  # // OBSOLETE Superseded by kCommandSetWindowBounds3.
    CommandSetTabIndexInWindow = 2
    CommandTabNavigationPathPrunedFromBack = 5  # // OBSOLETE: Superseded by kCommandTabNavigationPathPruned instead
    CommandUpdateTabNavigation = 6
    CommandSetSelectedNavigationIndex = 7
    CommandSetSelectedTabInIndex = 8
    CommandSetWindowType = 9
    CommandSetWindowBounds2 = 10  # // OBSOLETE Superseded by kCommandSetWindowBounds3. Except for data migration.
    CommandTabNavigationPathPrunedFromFront = 11  # // Superseded kCommandTabNavigationPathPruned instead
    CommandSetPinnedState = 12
    CommandSetExtensionAppID = 13
    CommandSetWindowBounds3 = 14
    CommandSetWindowAppName = 15
    CommandTabClosed = 16
    CommandWindowClosed = 17
    CommandSetTabUserAgentOverride = 18  # // OBSOLETE: Superseded by kCommandSetTabUserAgentOverride2.
    CommandSessionStorageAssociated = 19
    CommandSetActiveWindow = 20
    CommandLastActiveTime = 21
    CommandSetWindowWorkspace = 22  # // OBSOLETE Superseded by kCommandSetWindowWorkspace2.
    CommandSetWindowWorkspace2 = 23
    CommandTabNavigationPathPruned = 24
    CommandSetTabGroup = 25
    CommandSetTabGroupMetadata = 26  # // OBSOLETE Superseded by kCommandSetTabGroupMetadata2.
    CommandSetTabGroupMetadata2 = 27
    CommandSetTabGuid = 28
    CommandSetTabUserAgentOverride2 = 29
    CommandSetTabData = 30
    CommandSetWindowUserTitle = 31
    CommandSetWindowVisibleOnAllWorkspaces = 32
    CommandAddTabExtraData = 33
    CommandAddWindowExtraData = 34

    # Edge has custom command types. These are what I have seen so far.
    # None of these types appear to be related to browsing data at the moment (typically only a few bytes long).
    EdgeCommandUnknown131 = 131
    EdgeCommandUnknown132 = 132

    UnusedCommand = 255


class PageTransition:
    # ui/base/page_transition_types.h
    _core_mask = 0xff
    _qualifier_mask = 0xffffff00
    _core_transitions = {
            0: "Link",
            1: "Typed",
            2: "AutoBookmark",
            3: "AutoSubframe",
            4: "ManualSubframe",
            5: "Generated",
            6: "AutoToplevel",
            7: "FormSubmit",
            8: "Reload",
            9: "Keyword",
            10: "KeywordGenerated"
    }
    _qualifiers = {
            0x00800000: "Blocked",
            0x01000000: "ForwardBack",
            0x02000000: "FromAddressBar",
            0x04000000: "HomePage",
            0x08000000: "FromApi",
            0x10000000: "ChainStart",
            0x20000000: "ChainEnd",
            0x40000000: "ClientRedirect",
            0x80000000: "ServerRedirect"
    }

    def __init__(self, value):
        self._value = value
        if value < 0:
            # signed to unsigned
            value += (0x80000000 * 2)
        self._core_transition = PageTransition._core_transitions[value & PageTransition._core_mask]
        self._qualifiers = []
        for flag in PageTransition._qualifiers:
            if (value & PageTransition._qualifier_mask) & flag > 0:
                self._qualifiers.append(PageTransition._qualifiers[flag])

    def __str__(self):
        return "; ".join([self._core_transition] + self._qualifiers)

    def __repr__(self):
        return "<ChromeTransition ({0}): {1})>".format(self._value, str(self))

    @property
    def core_transition(self) -> str:
        return self._core_transition

    @property
    def qualifiers(self) -> typing.Iterable[str]:
        yield from self._qualifiers

    @property
    def value(self):
        return self._value


class SnssError(Exception):
    ...


@dataclasses.dataclass(frozen=True)
class SessionCommand:
    offset: int
    id_type: typing.Union[SessionRestoreIdType, TabRestoreIdType]


@dataclasses.dataclass(frozen=True)
class NavigationEntry(SessionCommand):
    # components/sessions/core/serialized_navigation_entry.cc
    index: int
    url: str
    title: str
    page_state_raw: bytes  # replace with completed PageState object
    transition_type: PageTransition
    has_post_data: typing.Optional[bool] = None
    referrer_url: typing.Optional[str] = None
    original_request_url: typing.Optional[str] = None
    is_overriding_user_agent: typing.Optional[bool] = None
    timestamp: typing.Optional[datetime.datetime] = None
    http_status: typing.Optional[int] = None
    referrer_policy: typing.Optional[int] = None
    extended_map: typing.Optional[types.MappingProxyType] = None
    task_id: typing.Optional[int] = None
    parent_task_id: typing.Optional[int] = None
    root_task_id: typing.Optional[int] = None
    session_id: typing.Optional[int] = None

    @classmethod
    def from_pickle(
            cls, pickle, id_type: typing.Union[SessionRestoreIdType, TabRestoreIdType],
            offset: int, session_id: typing.Optional[int]=None) -> "NavigationEntry":
        index = pickle.read_int32()
        url = pickle.read_string()
        title = pickle.read_string16()
        page_state_length = pickle.read_int32()
        # https://source.chromium.org/chromium/chromium/src/+/main:third_party/blink/common/page_state/page_state_serialization.cc;drc=d1e1301c82bef37d30796e2d6098856b851d90a4;l=897
        page_state_raw = pickle.read_aligned(page_state_length)
        transition_type = PageTransition(pickle.read_uint32())

        try:
            type_mask = pickle.read_uint32()
        except EasyPickleException:
            # very old versions of data end here, so we return a partial object here
            return cls(offset, id_type, index, url, title, page_state_raw, transition_type)

        has_post_data = (type_mask & 0x01) > 0
        referrer_url = pickle.read_string()
        _ = pickle.read_int32()  # referrer policy, not used
        original_request_url = pickle.read_string()
        is_overriding_user_agent = pickle.read_bool()
        timestamp = pickle.read_datetime()
        _ = pickle.read_string16()  # search terms, not used
        http_status = pickle.read_int32()
        referrer_policy = pickle.read_int32()

        extended_map_size = pickle.read_int32()
        extended_map = {}
        for _ in range(extended_map_size):
            key = pickle.read_string()
            value = pickle.read_string()
            extended_map[key] = value

        extended_map = types.MappingProxyType(extended_map)

        task_id = None
        parent_task_id = None
        root_task_id = None

        try:
            # these might not exist in older files, so no big deal if we can't get them
            task_id = pickle.read_int64()
            parent_task_id = pickle.read_int64()
            root_task_id = pickle.read_int64()

            child_task_id_count = pickle.read_int32()
            if child_task_id_count != 0:
                raise SnssError("Child tasks should not be present when reading NavigationEntry")
        except EasyPickleException:
            pass

        return cls(
            offset, id_type,
            index, url, title, page_state_raw, transition_type, has_post_data, referrer_url, original_request_url,
            is_overriding_user_agent, timestamp, http_status, referrer_policy, extended_map, task_id, parent_task_id,
            root_task_id, session_id
        )


class UnprocessedEntry(SessionCommand):
    ...


class SnssFileType(enum.Enum):
    Session = 1
    Tab = 2


class SnssFile:
    def __init__(self, file_type: SnssFileType, stream: typing.BinaryIO):
        # components/sessions/core/command_storage_backend.cc
        self._f = stream
        if file_type == SnssFileType.Session:
            self._id_type = SessionRestoreIdType
        elif file_type == SnssFileType.Tab:
            self._id_type = TabRestoreIdType
        else:
            raise ValueError("file_type is an unknown SnssFileType or is not SnssFileType")

        self._file_type = file_type
        header = self._f.read(8)
        if header[0:4] != b"SNSS":
            raise SnssError(f"Invalid magic; expected SNSS; got {header[0:4]}")
        self._version, = struct.unpack("<I", header[4:8])
        if self._version not in (1, 3):
            raise SnssError(f"Expected version 1 or 3, got version {self._version}")

    @property
    def file_type(self) -> SnssFileType:
        return self._file_type

    def reset(self):
        self._f.seek(8, os.SEEK_SET)

    def _get_next_session_command(self) -> typing.Optional[SessionCommand]:
        # components/sessions/core/command_storage_backend.cc
        start_offset = self._f.tell()
        length_raw = self._f.read(2)
        if not length_raw:
            return None  # eof
        length, = struct.unpack("<H", length_raw)
        data = self._f.read(length)
        if len(data) != length:
            raise ValueError(f"Could not get enough data reading record starting at {start_offset}")
        record_id_type = self._id_type(data[0])

        # components/sessions/core/session_service_commands.cc, components/sessions/core/base_session_service_commands.cc
        # components/sessions/core/tab_restore_service_impl.cc
        if record_id_type in (SessionRestoreIdType.CommandUpdateTabNavigation, TabRestoreIdType.CommandUpdateTabNavigation):  # 6, 1
            with EasyPickleIterator(data[1:]) as pickle:
                session_id = pickle.read_int32()
                nav = NavigationEntry.from_pickle(pickle, record_id_type, start_offset, session_id)
                return nav
        else:
            return UnprocessedEntry(start_offset, record_id_type)

    def iter_session_commands(self) -> typing.Iterable[SessionCommand]:
        self.reset()
        while command := self._get_next_session_command():
            yield command


def main(args):
    in_path = pathlib.Path(args[0])
    if in_path.name.startswith("Session_"):
        file_type = SnssFileType.Session
    elif in_path.name.startswith("Tabs_"):
        file_type = SnssFileType.Tab
    else:
        raise ValueError("File name does not start with Session or Tabs")
    with in_path.open("rb") as f:
        snss_file = SnssFile(file_type, f)
        for command in snss_file.iter_session_commands():
            print(command)


if __name__ == '__main__':
    main(sys.argv[1:])



from __future__ import annotations

from typing import Any

import pytest

from linear_mcp_fast import local_handlers


class MiniReader:
    def __init__(self):
        self.teams = {
            "T1": {"id": "T1", "key": "DEV", "name": "Dev Team", "description": "desc"},
        }
        self.users = {
            "U1": {"id": "U1", "name": "Alice", "displayName": "Alice", "email": "a@test"},
            "U2": {"id": "U2", "name": "Bob", "displayName": "Bob", "email": "b@test"},
        }
        self.states = {
            "S1": {"id": "S1", "name": "In Progress", "type": "started", "teamId": "T1", "position": 1},
            "S2": {"id": "S2", "name": "Backlog", "type": "backlog", "teamId": "T1", "position": 2},
        }
        self.projects = {
            "P1": {"id": "P1", "name": "Platform", "slugId": "platform", "state": "active", "description": "desc"},
        }
        self.issues = {
            "I1": {
                "id": "I1",
                "identifier": "DEV-1",
                "title": "Fix bug",
                "priority": 1,
                "estimate": 2,
                "stateId": "S1",
                "assigneeId": "U1",
                "teamId": "T1",
                "projectId": "P1",
                "dueDate": None,
                "description": "Issue 1",
                "createdAt": "2025-01-01T00:00:00Z",
                "updatedAt": "2025-01-03T00:00:00Z",
            },
            "I2": {
                "id": "I2",
                "identifier": "DEV-2",
                "title": "Write docs",
                "priority": 3,
                "estimate": 1,
                "stateId": "S2",
                "assigneeId": "U2",
                "teamId": "T1",
                "projectId": "P1",
                "dueDate": None,
                "description": "Issue 2",
                "createdAt": "2025-01-02T00:00:00Z",
                "updatedAt": "2025-01-02T00:00:00Z",
            },
        }
        self.labels = {"L1": {"id": "L1", "name": "bug", "color": "red", "isGroup": False, "teamId": "T1"}}
        self.initiatives = {
            "N1": {"id": "N1", "name": "North Star", "slugId": "north", "ownerId": "U1", "status": "active"}
        }
        self.cycles = {"C1": {"id": "C1", "number": 10, "teamId": "T1", "currentProgress": {"scopeCount": 2}}}
        self.documents = {
            "D1": {
                "id": "D1",
                "title": "Platform RFC",
                "slugId": "platform-rfc",
                "projectId": "P1",
                "creatorId": "U1",
                "createdAt": "2025-01-01",
                "updatedAt": "2025-01-05",
            }
        }
        self.milestones = {
            "M1": {
                "id": "M1",
                "name": "Alpha",
                "projectId": "P1",
                "sortOrder": 1,
                "targetDate": "2025-02-01",
                "currentProgress": {"completedIssueCount": 1, "scopeCount": 2},
            }
        }
        self.project_updates = {
            "UP1": {
                "id": "UP1",
                "body": "All good",
                "health": "onTrack",
                "projectId": "P1",
                "userId": "U1",
                "createdAt": "2025-01-04",
                "updatedAt": "2025-01-04",
            }
        }
        self._comments = {
            "I1": [{"id": "CMT1", "userId": "U1", "body": "LGTM", "createdAt": "2025-01-03", "updatedAt": "2025-01-03"}]
        }

    def _count(self, key: str, value: str | None) -> int:
        if not value:
            return 0
        return sum(1 for issue in self.issues.values() if issue.get(key) == value)

    def _state_counts(self, key: str, value: str | None) -> dict[str, int]:
        if not value:
            return {}
        out: dict[str, int] = {}
        for issue in self.issues.values():
            if issue.get(key) != value:
                continue
            state_type = self.get_state_type(issue.get("stateId", ""))
            out[state_type] = out.get(state_type, 0) + 1
        return out

    def find_user(self, search: str) -> dict[str, Any] | None:
        for user in self.users.values():
            if search.lower() in user["name"].lower():
                return user
        return None

    def find_team(self, search: str) -> dict[str, Any] | None:
        for team in self.teams.values():
            if search.upper() == team["key"] or search.lower() in team["name"].lower():
                return team
        return None

    def find_project(self, search: str) -> dict[str, Any] | None:
        for project in self.projects.values():
            if search.lower() in project["name"].lower() or search.lower() == project["slugId"]:
                return project
        return None

    def find_issue_status(self, team_id: str, query: str) -> dict[str, Any] | None:
        for state in self.states.values():
            if state.get("teamId") == team_id and query.lower() in state["name"].lower():
                return state
        return None

    def get_issue_by_identifier(self, identifier: str) -> dict[str, Any] | None:
        ident = identifier.upper()
        for issue in self.issues.values():
            if issue["identifier"].upper() == ident:
                return issue
        return None

    def get_comments_for_issue(self, issue_id: str) -> list[dict[str, Any]]:
        return list(self._comments.get(issue_id, []))

    def get_state_name(self, state_id: str) -> str:
        return self.states.get(state_id, {}).get("name", "Unknown")

    def get_state_type(self, state_id: str) -> str:
        return self.states.get(state_id, {}).get("type", "unknown")

    def get_user_name(self, user_id: str | None) -> str:
        if not user_id:
            return "Unassigned"
        return self.users.get(user_id, {}).get("name", "Unknown")

    def get_project_name(self, project_id: str | None) -> str:
        if not project_id:
            return ""
        return self.projects.get(project_id, {}).get("name", "")

    def get_issue_count_for_team(self, team_id: str | None) -> int:
        return self._count("teamId", team_id)

    def get_issue_count_for_project(self, project_id: str | None) -> int:
        return self._count("projectId", project_id)

    def get_issue_count_for_user(self, user_id: str | None) -> int:
        return self._count("assigneeId", user_id)

    def get_issue_state_counts_for_team(self, team_id: str | None) -> dict[str, int]:
        return self._state_counts("teamId", team_id)

    def get_issue_state_counts_for_project(self, project_id: str | None) -> dict[str, int]:
        return self._state_counts("projectId", project_id)

    def get_issue_state_counts_for_user(self, user_id: str | None) -> dict[str, int]:
        return self._state_counts("assigneeId", user_id)

    def get_cycles_for_team(self, team_id: str) -> list[dict[str, Any]]:
        return [c for c in self.cycles.values() if c.get("teamId") == team_id]

    def find_milestone(self, project_id: str, query: str) -> dict[str, Any] | None:
        for m in self.milestones.values():
            if m["projectId"] == project_id and query.lower() in m["name"].lower():
                return m
        return None

    def get_milestones_for_project(self, project_id: str) -> list[dict[str, Any]]:
        return [m for m in self.milestones.values() if m.get("projectId") == project_id]

    def find_initiative(self, query: str) -> dict[str, Any] | None:
        for initiative in self.initiatives.values():
            if query.lower() in initiative["name"].lower():
                return initiative
        return None

    def find_document(self, query: str) -> dict[str, Any] | None:
        for doc in self.documents.values():
            if query.lower() in doc["title"].lower() or query.lower() == doc["slugId"]:
                return doc
        return None


@pytest.fixture()
def reader() -> MiniReader:
    return MiniReader()


def test_list_issues_filters_and_limit(reader: MiniReader):
    result = local_handlers.list_issues(reader, assignee="Alice", limit=1, orderBy="updatedAt")
    assert result["totalCount"] == 1
    assert result["issues"][0]["identifier"] == "DEV-1"


def test_get_issue_returns_comments(reader: MiniReader):
    result = local_handlers.get_issue(reader, "DEV-1")
    assert result is not None
    assert result["comments"][0]["author"] == "Alice"


def test_get_team_uses_precomputed_counts(reader: MiniReader):
    result = local_handlers.get_team(reader, "DEV")
    assert result is not None
    assert result["issueCount"] == 2
    assert result["issuesByState"]["started"] == 1


def test_get_project_uses_find_project_and_state_counts(reader: MiniReader):
    result = local_handlers.get_project(reader, "platform")
    assert result is not None
    assert result["issueCount"] == 2
    assert result["issuesByState"]["backlog"] == 1


def test_get_status_updates_unsupported_filter_raises(reader: MiniReader):
    with pytest.raises(local_handlers.LocalFallbackRequested):
        local_handlers.get_status_updates(reader, type="project", initiative="north")


def test_get_status_updates_by_id(reader: MiniReader):
    result = local_handlers.get_status_updates(reader, type="project", id="UP1")
    assert result is not None
    assert result["id"] == "UP1"
    assert result["author"] == "Alice"


def test_get_document_by_slug(reader: MiniReader):
    result = local_handlers.get_document(reader, "platform-rfc")
    assert result is not None
    assert result["url"].endswith("/platform-rfc")


def test_list_milestones_serializes_progress(reader: MiniReader):
    result = local_handlers.list_milestones(reader, "Platform")
    assert len(result) == 1
    assert result[0]["progress"]["completed"] == 1



"""
Copyright 2021, CCL Forensics
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import pathlib
import datetime
import sqlite3
from ccl_chromium_reader import ccl_chromium_localstorage

__version__ = "0.1"
__description__ = "Dumps a Chromium localstorage leveldb to sqlite for review"
__contact__ = "Alex Caithness"

DB_SCHEMA = """
CREATE TABLE storage_keys ("_id" INTEGER PRIMARY KEY AUTOINCREMENT, "storage_key" TEXT);
CREATE TABLE batches ("start_ldbseq" INTEGER PRIMARY KEY, 
                      "end_ldbseq" INTEGER,
                      "storage_key" INTEGER,
                      "timestamp" INTEGER);
CREATE TABLE records ("_id" INTEGER PRIMARY KEY AUTOINCREMENT,
                      "storage_key" INTEGER,
                      "key" TEXT,
                      "value" TEXT,
                      "batch" INTEGER,
                      "ldbseq" INTEGER);
CREATE INDEX "storage_keys_storage_key" ON "storage_keys" ("storage_key");

CREATE VIEW "records_view" AS
    SELECT 
      storage_keys.storage_key AS "storage_key",
      records."key"  AS "key",
      records.value AS "value",
      datetime(batches."timestamp", 'unixepoch') AS "batch_timestamp",
      records.ldbseq AS "ldbseq"
    FROM records
      INNER JOIN storage_keys ON records.storage_key = storage_keys._id
      INNER JOIN batches ON records.batch = batches.start_ldbseq
    ORDER BY records.ldbseq;
"""

INSERT_STORAGE_KEY_SQL = """INSERT INTO "storage_keys" ("storage_key") VALUES (?);"""
INSERT_BATCH_SQL = """INSERT INTO "batches" ("start_ldbseq", "end_ldbseq", "storage_key", "timestamp") 
                      VALUES (?, ?, ?, ?);"""
INSERT_RECORD_SQL = """INSERT INTO "records" ("storage_key", "key", "value", "batch", "ldbseq")
                       VALUES (?, ?, ?, ?, ?);"""


def main(args):
    level_db_in_dir = pathlib.Path(args[0])
    db_out_path = pathlib.Path(args[1])

    if db_out_path.exists():
        raise ValueError("output database already exists")

    local_storage = ccl_chromium_localstorage.LocalStoreDb(level_db_in_dir)
    out_db = sqlite3.connect(db_out_path)
    out_db.executescript(DB_SCHEMA)
    cur = out_db.cursor()

    storage_keys_lookup = {}
    for storage_key in local_storage.iter_storage_keys():
        cur.execute(INSERT_STORAGE_KEY_SQL, (storage_key,))
        cur.execute("SELECT last_insert_rowid();")
        storage_key_id = cur.fetchone()[0]
        storage_keys_lookup[storage_key] = storage_key_id

    for batch in local_storage.iter_batches():
        cur.execute(
            INSERT_BATCH_SQL,
            (batch.start, batch.end, storage_keys_lookup[batch.storage_key],
             batch.timestamp.replace(tzinfo=datetime.timezone.utc).timestamp()))

    for record in local_storage.iter_all_records():
        batch = local_storage.find_batch(record.leveldb_seq_number)
        batch_id = batch.start if batch is not None else None
        cur.execute(
            INSERT_RECORD_SQL,
            (
                storage_keys_lookup[record.storage_key], record.script_key, record.value,
                batch_id, record.leveldb_seq_number
            )
        )

    cur.close()
    out_db.commit()
    out_db.close()


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(f"{pathlib.Path(sys.argv[0]).name} <leveldb dir> <out.db>")
        exit(1)
    main(sys.argv[1:])
